#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#

from copy import deepcopy
from collections import defaultdict, OrderedDict
from dataclasses import dataclass
import json
import warnings
import sys

import numpy as np
from rdkit import Chem
from rdkit.Chem import rdPartialCharges

from .utils import rdkitutils
from .utils import utils
from .utils.geomutils import calcDihedral
from .receptor_pdbqt import PDBQTReceptor


try:
    from openbabel import openbabel as ob
    from .utils import obutils
except ImportError:
    _has_openbabel = False
else:
    _has_openbabel = True

try:
    from misctools import StereoIsomorphism
except ImportError as _import_misctools_error:
    _has_misctools = False
else:
    _has_misctools = True
    

# TODO modify so that there are no more dictionaries and only list/arrays (fix me)
# methods like add_x,del_x,get_x will deal with indexing (fix me)
# the goal is to remove dictionaries (fix me)

# (fix me)) provide infrastructure to calculate "differential" modifications,
# like tautomers: the class calculating tautomers will copy and modify the setup
# to swap protons and change the atom types (and possibly rotations?), then
# there will e a function to extract only differences (i.e., atom types and coordinates,
# in the case of tautomers) and store them to make it a switch-state

# TODO use only 1D arrays (C)


# TODO update add_atom to not to specify neighbors, which should be defined only when a bond is created?
# TODO change all attributes_to_copy to have underscore ?

class MoleculeSetup:
    """ mol: molecule structurally prepared with explicit hydrogens

        the setup provides:
            - data storage
            - SMARTS matcher for all atom typers
    """

    def __init__(self):

        self.atom_pseudo = []
        self.coord = OrderedDict()  # FIXME all OrderedDict shuold be converted to lists?
        self.charge = OrderedDict()
        self.pdbinfo = OrderedDict()
        self.atom_type = OrderedDict()
        self.atom_params = {}
        self.dihedral_interactions = list()
        self.dihedral_partaking_atoms = dict()
        self.dihedral_labels = dict()
        self.atom_ignore = OrderedDict()
        self.chiral = OrderedDict()
        self.atom_true_count = 0
        self.graph = OrderedDict()
        self.bond = OrderedDict()
        self.element = OrderedDict()
        self.interaction_vector = OrderedDict()
        self.flexibility_model = {}
        self.ring_closure_info = {
            "bonds_removed": [],
            "pseudos_by_atom": {},
        }
        self.restraints = []
        self.is_sidechain = False
        self.rmsd_symmetry_indices = ()
        self.rings = {}
        self.rings_aromatic = []
        self.atom_to_ring_id = defaultdict(list)
        self.ring_corners = {}  # used to store corner flexibility
        self.name = None
        self.rotamers = []

    def copy(self):
        newsetup = MoleculeSetup()
        newsetup.__dict__ = deepcopy(self.__dict__)
        return newsetup

    @classmethod
    def from_mol(cls, mol, keep_chorded_rings=False, keep_equivalent_rings=False,
                 assign_charges=True):
        molsetup = cls()
        molsetup.mol = mol
        molsetup.atom_true_count = molsetup.get_num_mol_atoms()
        molsetup.name = molsetup.get_mol_name()
        molsetup.init_atom(assign_charges)
        molsetup.init_bond()
        molsetup.perceive_rings(keep_chorded_rings, keep_equivalent_rings)
        return molsetup

    @classmethod
    def from_prmtop_inpcrd(cls, prmtop, crd_filename):
        molsetup = cls()
        x, y, z = prmtop._coords_from_inpcrd(crd_filename)
        all_atom_indices = prmtop.vmdsel("")
        if len(all_atom_indices) != len(x):
            raise RuntimeError("number of prmtop atoms differs from x coordinates")
        prmtop.upper_atom_types = prmtop._gen_gaff_uppercase()
        charges = [chrg/18.2223 for chrg in prmtop.read_flag("CHARGE")]
        atomic_nrs = prmtop.read_flag("ATOMIC_NUMBER")
        pdbqt_string = ""
        molsetup.atom_params["rmin_half"] = []
        molsetup.atom_params["epsilon"] = []
        vdw_by_atype = {}
        for vdw in prmtop._retrieve_vdw():
            if vdw.atom in vdw_by_atype:
                raise RuntimeError("repeated atom type from prmtop._retrieve_vdw()")
            vdw_by_atype[vdw.atom] = {"rmin_half": vdw.r, "epsilon": vdw.e}
        for i in prmtop.atom_sel_idx:
            pdbinfo = rdkitutils.PDBAtomInfo(
                name=prmtop.pdb_atom_name[i],
                resName = prmtop.pdb_resname[i],
                resNum = int(prmtop.pdb_resid[i]),
                chain = " ",
            )
            atype = prmtop.upper_atom_types[i]
            molsetup.add_atom(
                i,
                coord = (x[i], y[i], z[i]),
                charge = charges[i],
                atom_type = atype,
                element = atomic_nrs[i], 
                pdbinfo = pdbinfo,
                chiral = False,
                ignore = False,
                add_atom_params = {
                    "rmin_half": vdw_by_atype[atype]["rmin_half"],
                    "epsilon": vdw_by_atype[atype]["epsilon"],
                }
            )
            #molsetup.atom_params["rmin_half"].append(vdw_by_atype[atype]["rmin_half"])
            #molsetup.atom_params["epsilon"].append(vdw_by_atype[atype]["epsilon"])
        bond_order = 1 # the prmtop does not have bond orders, so we set all to 1
        bonds_inc_h = prmtop.read_flag("BONDS_INC_HYDROGEN")
        bonds_not_h = prmtop.read_flag("BONDS_WITHOUT_HYDROGEN")
        bonds = bonds_inc_h + bonds_not_h
        nr_bonds = int(len(bonds) / 3) 
        for index_bond in range(nr_bonds):
            i_atom = int(bonds[index_bond * 3 + 0] / 3)
            j_atom = int(bonds[index_bond * 3 + 1] / 3)
            if (i_atom in prmtop.atom_sel_idx) and (j_atom in prmtop.atom_sel_idx):
                molsetup.add_bond(i_atom, j_atom, order=bond_order, rotatable=False)
        return molsetup


    @classmethod
    def from_pdb_and_residue_params(cls, pdb_fname, residue_params_fname=None):
        if residue_params_fname is not None:
            with open(residue_params_fname) as f:
                residue_params = json.load(f)
        with open(pdb_fname) as f:
            lines = f.readlines()
        x = []
        y = []
        z = []
        res_list = []
        atom_names_list = []
        last_res = None
        pdbinfos = []
        for line in lines:
            if not (line.startswith("ATOM") or line.startswith("HETATM")):
                continue
            resn = line[17:20]
            resi = int(line[22:26])
            chain = line[21:22]
            res = (resn, resi, chain)
            if res != last_res:
                res_list.append(resn)
                atom_names_list.append([])
            last_res = res
            atom_name = line[12:16].strip() 
            atom_names_list[-1].append(atom_name)
            x.append(float(line[30:38]))
            y.append(float(line[38:46]))
            z.append(float(line[46:54]))
            pdbinfo = rdkitutils.PDBAtomInfo(
                name=atom_name,
                resName=resn,
                resNum=resi,
                chain=chain
            )
            pdbinfos.append(pdbinfo)

        atom_params = {}
        atom_counter = 0
        all_ok = True
        all_err = ""
        for (res, atom_names) in zip(res_list, atom_names_list):
            p, ok, err = PDBQTReceptor.get_params_for_residue(res, atom_names)
            nr_params_to_add = set([len(values) for _, values in p.items()])
            if len(nr_params_to_add) != 1:
                raise RuntimeError("inconsistent number of parameters in %s" % p)
            nr_params_to_add = nr_params_to_add.pop()
            all_ok &= ok
            all_err += err
            for key in p:
                atom_params.setdefault(key, [None]*atom_counter)
                atom_params[key].extend(p[key])
            atom_counter += nr_params_to_add
            for key in atom_params:
                if key not in p:
                    atom_params[key].extend([None]*nr_params_to_add)
    
        if not all_ok:
            raise RuntimeError(all_err)
        
        molsetup = cls()
        charges = atom_params.pop("gasteiger")
        atypes = atom_params.pop("atom_types")
        for i in range(len(x)):
            molsetup.add_atom(
                i,
                coord = (x[i], y[i], z[i]),
                charge = charges[i],
                atom_type = atypes[i],
                element = None,
                pdbinfo = pdbinfo,
                chiral = False,
                ignore = False,
            )

        molsetup.atom_params = atom_params

        return molsetup 

    def set_types_from_uniq_atom_params(self, uniq_atom_params, prefix):
        param_idxs = uniq_atom_params.get_indices_from_atom_params(self.atom_params)
        if len(param_idxs) != len(self.atom_type):
            raise RuntimeError("{len(param_idxs)} parameters but {len(self.atom_type)} in molsetup")
        for i, j in enumerate(param_idxs):
            self.atom_type[i] = f"{prefix}{j}"
        return None


    def add_atom(self, idx=None, coord=np.array([0.0, 0.0,0.0], dtype='float'),
            element=None, charge=0.0, atom_type=None, pdbinfo=None,
            ignore=False, chiral=False, overwrite=False, add_atom_params=None):
        """ function to add all atom information at once;
            every property is optional
        """
        if idx is None:
            idx = len(self.coord)
        if idx in self.coord and not overwrite:
            print("ADD_ATOM> Error: the idx [%d] is already occupied (use 'overwrite' to force)")
            return False
        self.set_coord(idx, coord)
        self.set_charge(idx, charge)
        self.set_element(idx, element)
        self.set_atom_type(idx, atom_type)
        self.set_pdbinfo(idx, pdbinfo)
        self.set_chiral(idx, chiral)
        self.set_ignore(idx, ignore)
        self.graph.setdefault(idx, [])
        for key in self.atom_params:
            if type(add_atom_params) == dict and key in add_atom_params:
                value = add_atom_params[key]
            else:
                value = None
            self.atom_params[key].append(value)
        return idx

    def del_atom(self, idx):
        """ remove an atom and update all data associate with it """
        pass
        # coords
        # charge
        # element
        # type
        # neighbor graph
        # ignore
        # update bonds bonds (using the neighbor graph)
        # If pseudo-atom, update other information, too


    # pseudo-atoms
    def add_pseudo(self, coord=np.array([0.0,0.0,0.0], dtype='float'), charge=0.0,
            anchor_list=None, atom_type=None, rotatable=False,
            pdbinfo=None, directional_vectors=None, ignore=False, overwrite=False):
        """ add a new pseudoatom
            multiple bonds can be specified in "anchor_list" to support the centroids of aromatic rings

            if rotatable, makes the anchor atom rotatable to allow the pseudoatom movement
        """
        idx = self.atom_true_count + len(self.atom_pseudo)
        if idx in self.coord and not overwrite:
            print("ADD_PSEUDO> Error: the idx [%d] is already occupied (use 'overwrite' to force)")
            return False
        self.atom_pseudo.append(idx)
        # add the pseudoatom information to the atoms
        self.add_atom(idx=idx, coord=coord,
                element=0,
                charge=charge,
                atom_type=atom_type,
                pdbinfo=pdbinfo,
                ignore=ignore,
                overwrite=overwrite)
        # anchor atoms
        if not anchor_list is None:
            for anchor in anchor_list:
                self.add_bond(idx, anchor, order=0, rotatable=rotatable)
        # directional vectors
        if not directional_vectors is None:
            self.add_interaction_vector(idx, directional_vectors)
        return idx

    def add_bond(self, idx1, idx2, order=0, rotatable=False):

        if not idx2 in self.graph[idx1]:
            self.graph[idx1].append(idx2)
        if not idx1 in self.graph[idx2]:
            self.graph[idx2].append(idx1)
        self.set_bond(idx1, idx2, order, rotatable)

    def add_rotamer(self, indices_list, angles_list):
        xyz = self.coord
        rotamer = {}
        for (i1, i2, i3, i4), angle in zip(indices_list, angles_list):
            bond_id = self.get_bond_id(i2, i3)
            if bond_id in rotamer:
                raise RuntimeError("repeated bond %d-%d" % bond_id)
            if not self.bond[bond_id]["rotatable"]:
                raise RuntimeError("trying to add rotamer for non rotatable bond %d-%d" % bond_id)
            d0 = calcDihedral(xyz[i1], xyz[i2], xyz[i3], xyz[i4])
            rotamer[bond_id] = angle - d0
        self.rotamers.append(rotamer)

    def set_atom_type(self, idx, atom_type):
        """ set the atom type for atom index
        atom_type : int or str?
        return: void
        """
        self.atom_type[idx] = atom_type

    def get_atom_type(self, idx):
        """ return the atom type for atom index in the lookup table
        idx : int
        return: str
        """
        return self.atom_type[idx]

    # ignore flag
    def set_ignore(self, idx, state):
        """ set the ignore flag (bool) for the atom
        (used formerged hydrogen)
        idx: int
        state: bool
        """
        self.atom_ignore[idx] = state

    # charge
    def get_charge(self, idx):
        """ return partial charge for atom index
        idx: int

        """
        return self.charge[idx]

    def set_charge(self, idx, charge):
        """ set partial charge"""
        self.charge[idx] = charge

    def get_coord(self, idx):
        """ return coordinates of atom index"""
        return self.coord[idx]

    def set_coord(self, idx, coord):
        """ define coordinates of atom index"""
        self.coord[idx] = coord

    def get_neigh(self, idx):
        """ return atoms connected to atom index"""
        return self.graph[idx]

    def set_chiral(self, idx, chiral):
        """ set chiral flag for atom """
        self.chiral[idx] = chiral

    def get_chiral(self, idx):
        """ get chiral flag for atom """
        return self.chiral[idx]

    def get_ignore(self, idx):
        """ return if the atom is ignored"""
        return bool(self.atom_ignore[idx])

    def is_aromatic(self, idx):
        """ check if atom is aromatic """
        for r in self.rings_aromatic:
            if idx in r:
                return True
        return False

    def set_element(self, idx, elem_num):
        """ set the atomic number of the atom idx"""
        self.element[idx] = elem_num

    def get_element(self, idx):
        """ return the atomic number of the atom idx"""
        return self.element[idx]

    # def get_atom_ring_count(self, idx):
    #     """ return the number of rings to which this atom belongs"""
    #     # FIXME this should be replaced by self.get_atom_rings()
    #     return len(self.atom_to_ring_id[idx])

    def get_atom_rings(self, idx):
        # FIXME this should replace self.get_atom_ring_count()
        """ return the list of rings to which the atom idx belongs"""
        if idx in self.atom_to_ring_id:
            return self.atom_to_ring_id[idx]
        return []

    def get_atom_indices(self, true_atoms_only=False):
        """ return the indices of the atoms registered in the setup
            if 'true_atoms_only' are requested, then pseudoatoms are ignored
        """
        indices = list(self.coord.keys())
        if true_atoms_only:
            return [ x for x in indices if not x in self.atom_pseudo ]
        return indices

    # interaction vectors
    def add_interaction_vector(self, idx, vector_list):
        """ add vector list to list of directional interaction vectors for atom idx"""
        if not idx in self.interaction_vector:
            self.interaction_vector[idx] = []
        for vec in vector_list:
            self.interaction_vector[idx].append(vec)

    # TODO evaluate if useful
    def _get_attrib(self, idx, attrib, default=None):
        """ generic function to provide a default for retrieving properties and returning standard values """
        return getattr(self, attrib).get(idx, default)

    def get_interaction_vector(self, idx):
        """ get list of directional interaction vectors for atom idx"""
        return self.interaction_vector[idx]

    def del_interaction_vector(self, idx):
        """ delete list of directional interaction vectors for atom idx"""
        del self.interaction_vector[idx]

    def set_pdbinfo(self, idx, data):
        """ add PDB data (resname/num, atom name, etc.) to the atom """
        self.pdbinfo[idx] = data

    def get_pdbinfo(self, idx):
        """ retrieve PDB data (resname/num, atom name, etc.) to the atom """
        return self.pdbinfo[idx]

    def set_bond(self, idx1, idx2, order=0, rotatable=False):
        """ populate bond lookup table with properties
            bonds are identified by any tuple of atom indices
            the function generates the canonical bond id

            order      : int
            rotatable  : bool
        """
        bond_id = self.get_bond_id(idx1, idx2)
        self.bond[bond_id] = {
            'bond_order': order,
            'rotatable': rotatable,
        }

    def del_bond(self, idx1, idx2):
        """ remove a bond from the lookup table """
        bond_id = self.get_bond_id(idx1, idx2)
        del self.bond[bond_id]
        self.graph[idx1].remove(idx2)
        # TODO check if we want to delete nodes that have no connections (we might want to keep them)
        if not self.graph[idx1]:
            del self.graph[idx1]
        self.graph[idx2].remove(idx1)
        if not self.graph[idx2]:
            del self.graph[idx2]

    def get_bond(self, idx1, idx2):
        """ return properties of a bond in the lookup table
            if the bond does not exist, None is returned

            idx1, idx2 : int

            return: dict or voidko
        """
        bond_idx = self.get_bond_id(idx1, idx2)
        try:
            return self.bond[bond_idx]
        except IndexError:
            return None

    @staticmethod
    def get_bond_id(idx1, idx2):
        """ used to generate canonical bond id from a pair of nodes in the graph"""
        idx_min = min(idx1, idx2)
        idx_max = max(idx1, idx2)
        return (idx_min, idx_max)

    # replaced by
    # def ring_atom_to_ring(self, arg):
    #     return self.atom_to_ring_id[arg]

    def get_bonds_in_ring(self, ring):
        """ input: 'ring' (list of atom indices)
            returns list of bonds in ring, each bond is a pair of atom indices
        """
        n = len(ring)
        bonds = []
        for i in range(n):
            bond = (ring[i], ring[(i+1) % n])
            bond = (min(bond), max(bond))
            bonds.append(bond)
        return bonds

    def walk_recursive(self, idx, collected=None, exclude=None):
        """ walk molecular graph and return subgraphs that are bond-connected"""
        if collected is None:
            collected = []
        if exclude is None:
            exclude = []
        for neigh in self.get_neigh(idx):
            if neigh in exclude:
                continue
            collected.append(neigh)
            exclude.append(neigh)
            self.walk_recursive(neigh, collected, exclude)
        return collected

    def copy_attributes_from(self, template):
        """ copy attributes to duplicate the template setup
        NOTE: the molecule will always keep the original setup (i.e., template)"""
        # TODO enable some kind of plugin system here too, to allow other objects to
        # add attributes?
        # TODO although, this would make the setup more fragile-> better have attributes
        # explicitely added here, and that's it
        for attr in self.attributes_to_copy:
            attr_copy = deepcopy(getattr(template, attr))
            setattr(self, attr, attr_copy)
        # TODO possible BUG? the molecule is shared by the different setups
        #      if one of them alters the molecule, properties will not be the same
        self.mol = template.mol

    def merge_terminal_atoms(self, indices):
        """for merging hydrogens, but will merge any atom or pseudo atom
            that is bonded to only one other atom"""

        for index in indices:
            if len(self.graph[index]) != 1:
                msg = "Atempted to merge atom %d with %d neighbors. "
                msg += "Only atoms with one neighbor can be merged."
                msg = msg % (index + 1, len(self.graph[index]))
                raise RuntimeError(msg)
            neighbor_index = self.graph[index][0]
            self.charge[neighbor_index] += self.get_charge(index)
            self.charge[index] = 0.0
            self.set_ignore(index, True)
            
    def init_atom(self):
        """ iterate through molecule atoms and build the atoms table """
        raise NotImplementedError("This method must be overloaded by inheriting class")

    def perceive_rings(self, keep_chorded_rings, keep_equivalent_rings):
        """ populate the rings and aromatic rings tableshe atoms table:
        self.rings_aromatics : list
            contains the list of ring_id items that are aromatic
        self.rings: dict
            store information about rings, like if they have corners that can
            be flipped and the graph of atoms that belong to them:

                self.rings[ring_id] = {
                                'corner_flip': False
                                'graph': {}
                                }

            The atom is built using the `walk_recursive` method
        self.ring_atom_to_ring_id: dict
            mapping of each atom belonginig to the ring: atom_idx -> ring_id
        """

        def isRingAromatic(ring_atom_indices):
            for atom_idx1, atom_idx2 in self.get_bonds_in_ring(ring_atom_indices):
                bond = self.mol.GetBondBetweenAtoms(atom_idx1, atom_idx2)
                if not bond.GetIsAromatic():
                    return False
            return True

        hjk_ring_detection = utils.HJKRingDetection(self.graph) 
        rings = hjk_ring_detection.scan(keep_chorded_rings, keep_equivalent_rings) # list of tuples of atom indices
        for ring_atom_idxs in rings:
            if isRingAromatic(ring_atom_idxs):
                self.rings_aromatic.append(ring_atom_idxs)
            self.rings[ring_atom_idxs] = {'corner_flip':False}
            graph = {}
            for atom_idx in ring_atom_idxs:
                self.atom_to_ring_id[atom_idx].append(ring_atom_idxs)
                # graph of atoms affected by potential ring movements
                graph[atom_idx] = self.walk_recursive(atom_idx, collected=[], exclude=list(ring_atom_idxs))
            self.rings[ring_atom_idxs]['graph'] = graph

    def add_dihedral_interaction(self, fourier_series):
        """ """
        index = 0
        for existing_fs in self.dihedral_interactions:
            if self.are_fourier_series_identical(existing_fs, fourier_series):
                return index
            index += 1
        safe_copy = json.loads(json.dumps(fourier_series))
        self.dihedral_interactions.append(safe_copy)
        return index

    @staticmethod
    def are_fourier_series_identical(fs1, fs2):
        index_by_periodicity1 = {fs1[index]["periodicity"]: index for index in range(len(fs1))}
        index_by_periodicity2 = {fs2[index]["periodicity"]: index for index in range(len(fs2))}
        if index_by_periodicity1 != index_by_periodicity2:
            return False
        for periodicity in index_by_periodicity1:
            index1 = index_by_periodicity1[periodicity]
            index2 = index_by_periodicity2[periodicity]
            for key in ["k", "phase", "periodicity"]:
                if fs1[index1][key] != fs2[index2][key]:
                    return False
        return True

    def init_bond(self):
        """ iterate through molecule bonds and build the bond table (id, table)
            CALCULATE
                bond_order: int
                    0       : pseudo-bond (for pseudoatoms)
                    1,2,3   : single-triple bond
                    5       : aromatic
                    999     : rigid

                if bond is in ring (both start and end atom indices in the bond are in the same ring)

            SETUP OPERATION
                Setup.add_bond(idx1, idx2, order, rotatable)
        """
        raise NotImplementedError("This method must be overloaded by inheriting class")

    def get_mol_name(self):
        raise NotImplementedError("This method must be overloaded by inheriting class")

    def find_pattern(self, smarts):
        raise NotImplementedError("This method must be overloaded by inheriting class")

    def get_smiles_and_order(self):
        raise NotImplementedError("This method must be overloaded by inheriting class")

    def write_xyz_string(self):
        n = len(self.element)
        string = "%d\n\n" % n
        for index in range(n):
            if index < self.atom_true_count:
                element = utils.mini_periodic_table[self.element[index]]
            else:
                element = "Ne"
            x, y, z = self.coord[index]
            string += "%3s %12.6f %12.6f %12.6f\n" % (element, x, y, z)
        return string
            
    def show(self):
        tot_charge = 0

        print("Molecule setup\n")
        print("==============[ ATOMS ]===================================================")
        print("idx  |          coords            | charge |ign| atype    | connections")
        print("-----+----------------------------+--------+---+----------+--------------- . . . ")
        for k, v in list(self.coord.items()):
            print("% 4d | % 8.3f % 8.3f % 8.3f | % 1.3f | %d" % (k, v[0], v[1], v[2],
                  self.charge[k], self.atom_ignore[k]),
                  "| % -8s |" % self.atom_type[k],
                  self.graph[k])
            tot_charge += self.charge[k]
        print("-----+----------------------------+--------+---+----------+--------------- . . . ")
        print("  TOT CHARGE: %3.3f" % tot_charge)

        print("\n======[ DIRECTIONAL VECTORS ]==========")
        for k, v in list(self.coord.items()):
            if k in self.interaction_vector:
                print("% 4d " % k, self.atom_type[k], end=' ')

        print("\n==============[ BONDS ]================")
        # For sanity users, we won't show those keys for now
        keys_to_not_show = ['bond_order', 'type']
        for k, v in list(self.bond.items()):
            t = ', '.join('%s: %s' % (i, j) for i, j in v.items() if not i in keys_to_not_show)
            print("% 8s - " % str(k), t)

        # _macrocycle_typer.show_macrocycle_scores(self)

        print('')

class RDKitMoleculeSetup(MoleculeSetup):

    @classmethod
    def from_mol(cls, mol, keep_chorded_rings=False, keep_equivalent_rings=False,
                 assign_charges=True, conformer_id=-1):
        if cls.has_implicit_hydrogens(mol):
            raise ValueError("RDKit molecule has implicit Hs. Need explicit Hs.")
        if mol.GetNumConformers() == 0: 
            raise ValueError("RDKit molecule does not have a conformer. Need 3D coordinates.")
        rdkit_conformer = mol.GetConformer(conformer_id) 
        if not rdkit_conformer.Is3D():
            warnings.warn("RDKit molecule not labeled as 3D. This warning won't show again.")
            RDKitMoleculeSetup.warned_not3D = True
        if mol.GetNumConformers() > 1 and conformer_id == -1:
            msg = "RDKit molecule has multiple conformers. Considering only the first one." 
            print(msg, file=sys.stderr)
        molsetup = cls()
        molsetup.mol = mol
        molsetup.atom_true_count = molsetup.get_num_mol_atoms()
        molsetup.name = molsetup.get_mol_name()
        coords = rdkit_conformer.GetPositions()
        molsetup.init_atom(assign_charges, coords)
        molsetup.init_bond()
        molsetup.perceive_rings(keep_chorded_rings, keep_equivalent_rings)
        molsetup.rmsd_symmetry_indices = cls.get_symmetries_for_rmsd(mol)

        # to store sets of coordinates, e.g. docked poses, as dictionaries indexed by
        # the atom index, because not all atoms need to have new coordinates specified
        # Unspecified hydrogen positions bonded to modified heavy atom positions
        # are to be calculated "on-the-fly".
        molsetup.modified_atom_positions = [] # list of dictionaries where keys are atom indices

        return molsetup

    def get_conformer_with_modified_positions(self, new_atom_positions):
        # we operate on one conformer at a time because SetTerminalAtomPositions
        # acts on all conformers of a molecule, and we don't want to guarantee
        # that all conformers require the same set of terminal atoms to be updated
        new_mol = Chem.Mol(self.mol)
        new_conformer = Chem.Conformer(self.mol.GetConformer())
        is_set_list = [False] * self.mol.GetNumAtoms()
        for atom_index, new_position in new_atom_positions.items():
            new_conformer.SetAtomPosition(atom_index, new_position)
            is_set_list[atom_index] = True
        new_mol.RemoveAllConformers()
        new_mol.AddConformer(new_conformer, assignId=True)
        for atom_index, is_set in enumerate(is_set_list):
            if not is_set and new_mol.GetAtomWithIdx(atom_idx).GetAtomicNum() == 1:
                neighbors = new_mol.GetAtomWithIdx(atom_idx).GetNeighbors()
                if len(neighbors) != 1:
                    raise RuntimeError("Expected H to have one neighbors")
                Chem.SetTerminalAtomCoords(new_mol, atom_index, neighbors[0].GetIdx())
        return new_conformer

    def get_mol_with_modified_positions(self, new_atom_positions_list=None):
        if new_atom_positions_list is None:
            new_atom_positions_list = self.modified_atom_positions
        new_mol = Chem.Mol(self.mol)
        new_mol.RemoveAllConformers()
        for new_atom_positions in new_atom_positions_list:
            conformer = self.get_conformer_with_modified_positions(new_atom_positions)
            new_mol.AddConformer(conformer, assignId=True)
        return new_mol

    def get_smiles_and_order(self):
        """
            return the SMILES after Chem.RemoveHs()
            and the mapping between atom indices in smiles and self.mol
        """

        ### # support sidechains in which not all real atoms are included in PDBQT
        ### ignored = []
        ### for atom in self.mol.GetAtoms():
        ###     index = atom.GetIdx()
        ###     if self.atom_ignore[index]:
        ###         ignored.append(index)
        ### if len(ignored) > 0: # e.g. sidechain padded with amides
        ###     mol_no_ignore = Chem.EditableMol(self.mol)
        ###     for index in sorted(ignored, reverse=True):
        ###         mol_no_ignore.RemoveAtom(index)
        ###     # remove dangling Hs
        ###     dangling_hs = []
        ###     for atom in mol_no_ignore.GetMol().GetAtoms():
        ###         if atom.GetAtomicNum() == 1:
        ###             if len(atom.GetNeighbors()) == 0:
        ###                 dangling_hs.append(atom.GetIdx())
        ###     for index in sorted(dangling_hs, reverse=True):
        ###         mol_no_ignore.RemoveAtom(index)
        ###     mol_no_ignore = mol_no_ignore.GetMol()
        ###     Chem.SanitizeMol(mol_no_ignore) 
        ### else:
        mol_no_ignore = self.mol

        # 3D SDF files written by other toolkits (OEChem, ChemAxon)
        # seem to not include the chiral flag in the bonds block, only in
        # the atoms block. RDKit ignores the atoms chiral flag as per the
        # spec. When reading SDF (e.g. from PubChem/PDB),
        # we may need to have RDKit assign stereo from coordinates, see:
        # https://sourceforge.net/p/rdkit/mailman/message/34399371/
        ps = Chem.RemoveHsParameters()
        # a user reported PDBbind Mol Blocks to have hcount=1 for Hs,
        # which adds a query to the RDKit H atom and then Chem.RemoveHs
        # does not remove Hs with queries by default
        # https://github.com/forlilab/Meeko/issues/62
        # https://github.com/rdkit/rdkit/issues/6615
        ps.removeWithQuery = True 
        mol_noH = Chem.RemoveHs(mol_no_ignore, ps) # imines (=NH) may become chiral
        # stereo imines [H]/N=C keep [H] after RemoveHs()
        # H isotopes also kept after RemoveHs()
        atomic_num_mol_noH = [atom.GetAtomicNum() for atom in mol_noH.GetAtoms()]
        noH_to_H = []
        parents_of_hs = {}
        for (index, atom) in enumerate(mol_no_ignore.GetAtoms()):
            if atom.GetAtomicNum() == 1: continue
            for i in range(len(noH_to_H), len(atomic_num_mol_noH)):
                if atomic_num_mol_noH[i] > 1:
                    break
                h_atom = mol_noH.GetAtomWithIdx(len(noH_to_H))
                assert(h_atom.GetAtomicNum() == 1)
                neighbors = h_atom.GetNeighbors()
                assert(len(neighbors) == 1)
                parents_of_hs[len(noH_to_H)] = neighbors[0].GetIdx()
                noH_to_H.append('H')
            noH_to_H.append(index)
        extra_hydrogens = len(atomic_num_mol_noH) - len(noH_to_H)
        if extra_hydrogens > 0:
            assert(set(atomic_num_mol_noH[len(noH_to_H):]) == {1})
        for i in range(extra_hydrogens):
            h_atom = mol_noH.GetAtomWithIdx(len(noH_to_H))
            assert(h_atom.GetAtomicNum() == 1)
            neighbors = h_atom.GetNeighbors()
            assert(len(neighbors) == 1)
            parents_of_hs[len(noH_to_H)] = neighbors[0].GetIdx()
            noH_to_H.append('H')

        # noH_to_H has the same length as the number of atoms in mol_noH
        # and each value is:
        #    - the index of the corresponding atom in mol, if value is integer
        #    - an hydrogen, if value is "H"
        # now, we need to replace those "H" with integers
        # "H" occur with stereo imine (e.g. [H]/N=C) and heavy Hs (e.g. [2H])
        hs_by_parent = {}
        for hidx, pidx in parents_of_hs.items():
            hs_by_parent.setdefault(pidx, [])
            hs_by_parent[pidx].append(hidx)
        for pidx, hidxs in hs_by_parent.items():
            siblings_of_h = [atom for atom in mol_no_ignore.GetAtomWithIdx(noH_to_H[pidx]).GetNeighbors() if atom.GetAtomicNum() == 1]
            sortidx = [i for i, j in sorted(list(enumerate(siblings_of_h)), key=lambda x: x[1].GetIdx())]
            if len(hidxs) == len(siblings_of_h):  
                # This is the easy case, just map H to each other in the order they appear
                for i, hidx in enumerate(hidxs):
                    noH_to_H[hidx] = siblings_of_h[sortidx[i]].GetIdx()
            elif len(hidxs) < len(siblings_of_h):
                # check hydrogen isotopes
                sibling_isotopes = [siblings_of_h[sortidx[i]].GetIsotope() for i in range(len(siblings_of_h))]
                molnoH_isotopes = [mol_noH.GetAtomWithIdx(hidx) for hidx in hidxs]
                matches = []
                for i, sibling_isotope in enumerate(sibling_isotopes):
                    for hidx in hidxs[len(matches):]:
                        if mol_noH.GetAtomWithIdx(hidx).GetIsotope() == sibling_isotope:
                            matches.append(i)
                            break
                if len(matches) != len(hidxs):
                    raise RuntimeError("Number of matched isotopes %d differs from query Hs: %d" % (len(matched), len(hidxs)))
                for hidx, i in zip(hidxs, matches):
                    noH_to_H[hidx] = siblings_of_h[sortidx[i]].GetIdx()
            else:
                raise RuntimeError("nr of Hs in mol_noH bonded to an atom exceeds nr of Hs in mol_no_ignore")

        smiles = Chem.MolToSmiles(mol_noH)
        order_string = mol_noH.GetProp("_smilesAtomOutputOrder")
        order_string = order_string.replace(',]', ']') # remove trailing comma
        order = json.loads(order_string) # mol_noH to smiles
        order = list(np.argsort(order))
        order = {noH_to_H[i]: order[i]+1 for i in range(len(order))} # 1-index
        return smiles, order

    def find_pattern(self, smarts):
        p = Chem.MolFromSmarts(smarts)
        return self.mol.GetSubstructMatches(p)

    def get_mol_name(self):
        if self.mol.HasProp("_Name"):
            return self.mol.GetProp("_Name")
        else:
            return None

    def get_num_mol_atoms(self):
        return self.mol.GetNumAtoms()

    def get_equivalent_atoms(self):
       return list(Chem.CanonicalRankAtoms(self.mol, breakTies=False))

    @staticmethod
    def get_symmetries_for_rmsd(mol, max_matches=17):
        mol_noHs = Chem.RemoveHs(mol)
        matches = mol.GetSubstructMatches(mol_noHs, uniquify=False, maxMatches=max_matches)
        if len(matches) == max_matches:
            if mol.HasProp("_Name"):
                molname = mol.GetProp("_Name")
            else:
                molname = ""
            print("warning: found the maximum nr of matches (%d) in RDKitMolSetup.get_symmetries_for_rmsd" % max_matches)
            print("Maybe this molecule is \"too\" symmetric? %s" % molname, Chem.MolToSmiles(mol_noHs))
        return matches

    def init_atom(self, assign_charges, coords):
        """ initialize the atom table information """
        # extract/generate charges
        if assign_charges:
            copy_mol = Chem.Mol(self.mol)
            for atom in copy_mol.GetAtoms():
                if atom.GetAtomicNum() == 34:
                    atom.SetAtomicNum(16)
            rdPartialCharges.ComputeGasteigerCharges(copy_mol)
            charges = [a.GetDoubleProp('_GasteigerCharge') for a in copy_mol.GetAtoms()]
        else:
            charges = [0.0] * self.mol.GetNumAtoms()
        # perceive chirality
        # TODO check consistency for chiral model between OB and RDKit
        chiral_info = {}
        for data in Chem.FindMolChiralCenters(self.mol, includeUnassigned=True):
            chiral_info[data[0]] = data[1]
        # register atom
        for a in self.mol.GetAtoms():
            idx = a.GetIdx()
            chiral = False
            if idx in chiral_info:
                chiral = chiral_info[idx]
            self.add_atom(idx,
                    coord=coords[idx],
                    element=a.GetAtomicNum(),
                    charge=charges[idx],
                    atom_type=None,
                    pdbinfo = rdkitutils.getPdbInfoNoNull(a),
                    chiral=False,
                    ignore=False)

    def init_bond(self):
        """ initialize bond information """
        for b in self.mol.GetBonds():
            idx1 = b.GetBeginAtomIdx()
            idx2 = b.GetEndAtomIdx()
            bond_order = int(b.GetBondType())
            # fix the RDKit aromatic type (FIXME)
            if bond_order == 12: # aromatic
                bond_order = 5
            if bond_order == 1:
                rotatable = True
            else:
                rotatable = False
            self.add_bond(idx1, idx2, order=bond_order, rotatable=rotatable)

    def copy(self):
        """ return a copy of the current setup"""
        newsetup = RDKitMoleculeSetup()
        for key, value in self.__dict__.items():
            if key != "mol":
                newsetup.__dict__[key] = deepcopy(value)
        newsetup.mol = Chem.Mol(self.mol) # not sure how deep of a copy this is
        return newsetup

    @staticmethod
    def has_implicit_hydrogens(mol):
        # based on needsHs from RDKit's AddHs.cpp
        for atom in mol.GetAtoms():
            nr_H_neighbors = 0
            for neighbor in atom.GetNeighbors():
                nr_H_neighbors += int(neighbor.GetAtomicNum() == 1)
            if atom.GetTotalNumHs(includeNeighbors=False) > nr_H_neighbors:
                return True
        return False

    def restrain_to(self, target_mol, kcal_per_angstrom_square=1.0, delay_angstroms=2.0):
        if not _has_misctools:
            raise ImportError(_import_misctools_error)
        stereo_isomorphism = StereoIsomorphism()
        mapping, idx = stereo_isomorphism(target_mol, self.mol)
        lig_to_drive = {b: a for (a, b) in mapping}
        num_real_atoms = target_mol.GetNumAtoms()
        target_positions = target_mol.GetConformer().GetPositions()
        for atom_index in range(len(mapping)):
            target_xyz = target_positions[lig_to_drive[atom_index]]
            restraint = Restraint(atom_index, target_xyz, kcal_per_angstrom_square, delay_angstroms)
            self.restraints.append(restraint)
        return


class OBMoleculeSetup(MoleculeSetup):

    def get_mol_name(self):
        return self.mol.GetTitle()

    def find_pattern(self, smarts):
        obsmarts = ob.OBSmartsPattern()
        obsmarts.Init(smarts)
        found = obsmarts.Match(self.mol)
        output = []
        if found:
            for x in obsmarts.GetUMapList():
                output.append([y-1 for y in x])
        return output

    def get_num_mol_atoms(self):
        return self.mol.NumAtoms()

    def get_equivalent_atoms(self):
        raise NotImplementedError 
    
    def init_atom(self, assign_charges):
        """initialize atom data table"""
        for a in ob.OBMolAtomIter(self.mol):
            partial_charge = a.GetPartialCharge() * float(assign_charges)
            self.add_atom(a.GetIdx() - 1,
                    coord=np.asarray(obutils.getAtomCoords(a), dtype='float'),
                    element=a.GetAtomicNum(),
                    charge=partial_charge,
                    atom_type=None,
                    pdbinfo = obutils.getPdbInfoNoNull(a),
                    ignore=False, chiral=a.IsChiral())
            # TODO check consistency for chiral model between OB and RDKit

    def init_bond(self):
        """initialize bond data table"""
        for b in ob.OBMolBondIter(self.mol):
            idx1 = b.GetBeginAtomIdx() - 1
            idx2 = b.GetEndAtomIdx() - 1
            bond_order = b.GetBondOrder()
            if b.IsAromatic():
                bond_order = 5
            self.add_bond(idx1, idx2, order=bond_order)

    def copy(self):
        """ return a copy of the current setup"""
        return OBMoleculeSetup(template=self)


class UniqAtomParams:
    def __init__(self):
        self.params = [] # aka rows
        self.param_names = [] # aka column names

    @classmethod
    def from_dict(cls, dictionary):
        uap = UniqAtomParams()
        uap.params = [row.copy() for row in dictionary["params"]]
        uap.param_names = dictionary["param_names"].copy()
        return uap

    def get_indices_from_atom_params(self, atom_params):
        nr_items = set([len(values) for key, values in atom_params.items()])
        if len(nr_items) != 1:
            raise RuntimeError(f"all lists in atom_params must have same length, got {nr_items}")
        if set(atom_params) != set(self.param_names):
            msg = f"parameter names in atom_params differ from internal ones\n"
            msg += f"  - in atom_params: {set(atom_params)}"
            msg += f"  - internal: {set(self.param_names)}"
            raise RuntimeError(msg)
        nr_items = nr_items.pop()
        param_idxs = []
        for i in range(nr_items):
            row = [atom_params[key][i] for key in self.param_names]
            param_index = None
            for j, existing_row in enumerate(self.params):
                if row == existing_row:
                    param_index = j
                    break
            param_idxs.append(param_index)
        return param_idxs 

    def add_parameter(self, new_param_dict):
        # remove None values to avoid a column with only Nones
        new_param_dict = {k: v for k, v in new_param_dict.items() if v is not None}
        incoming_keys = set(new_param_dict.keys())
        existing_keys = set(self.param_names)
        new_keys = incoming_keys.difference(existing_keys)
        for new_key in new_keys:
            self.param_names.append(new_key)
            for row in self.params:
                row.append(None) # fill in empty "cell" in new "column"

        new_row = []
        for key in self.param_names:
            value = new_param_dict.get(key, None)
            new_row.append(value)

        if len(new_keys) == 0: # try to match with existing row
            for index, row in enumerate(self.params):
                if row == new_row:
                    return index

        # if we are here, we didn't match
        new_row_index = len(self.params)
        self.params.append(new_row)
        return new_row_index


    def add_molsetup(self, molsetup, add_atomic_nr=False, add_atom_type=False):
        if "q" in molsetup.atom_params or "atom_type" in molsetup.atom_params:
            msg = '"q" and "atom_type" found in molsetup.atom_params'
            msg += ' but are hard-coded to store molsetup.charge and'
            msg += ' molsetup.atom_type in the internal data structure'
            raise RuntimeError(msg)
        param_idxs = []
        for atom_index, ignore in molsetup.atom_ignore.items():
            if ignore:
                param_idx = None
            else:
                p = {k: v[atom_index] for (k, v) in molsetup.atom_params.items()}
                if add_atomic_nr:
                    if "atomic_nr" in p:
                        raise RuntimeError("trying to add atomic_nr but it's already in atom_params")
                    p["atomic_nr"] = molsetup.element[atom_index]
                if add_atom_type:
                    if "atom_type" in p:
                        raise RuntimeError("trying to add atom_type but it's already in atom_params")
                    p["atom_type"] = molsetup.atom_type[atom_index]
                param_idx = self.add_parameter(p)
            param_idxs.append(param_idx)
        return param_idxs


@dataclass
class Restraint:
    atom_index: int
    target_xyz: (float, float, float)
    kcal_per_angstrom_square: float
    delay_angstroms: float

    def copy(self):
        new_target_xyz = (
            self.target_xyz[0],
            self.target_xyz[1],
            self.target_xyz[2])
        new_restraint = Restraint(
            self.atom_index,
            new_target_xyz,
            self.kcal_per_angstrom_square,
            self.delay_angstroms)
        return new_restraint



import pathlib
import json
from rdkit import Chem
from rdkit.Chem import rdFMCS
from rdkit.Chem import rdChemReactions
from rdkit.Chem.AllChem import EmbedMolecule, AssignBondOrdersFromTemplate
from .writer import PDBQTWriterLegacy
from .molsetup import RDKitMoleculeSetup
from .utils.rdkitutils import mini_periodic_table
from .utils.rdkitutils import react_and_map

import numpy as np

pkg_dir = pathlib.Path(__file__).parents[0]
with open(pkg_dir / "data" / "prot_res_params.json") as f:
    chorizo_params = json.load(f)

def mapping_by_mcs(mol, ref):
    mcs_result = rdFMCS.FindMCS([mol,ref], bondCompare=rdFMCS.BondCompare.CompareAny)
    mcs_mol = Chem.MolFromSmarts(mcs_result.smartsString)

    mol_idxs = mol.GetSubstructMatch(mcs_mol)
    ref_idxs = ref.GetSubstructMatch(mcs_mol)

    atom_map = {i:j for (i,j) in zip(mol_idxs, ref_idxs)}
    return atom_map

def reassign_formal_charge(mol, ref, mapping):
    #TODO this could be optimized
    #TODO ref charges could be precalculated to speed up large structures
    mol_charged_atoms = []
    for idx, atom in enumerate(mol.GetAtoms()):
        if atom.GetFormalCharge() != 0:
            mol_charged_atoms.append(idx)
            
    ref_charged_atoms = []
    for idx, atom in enumerate(ref.GetAtoms()):
        if atom.GetFormalCharge() != 0:
            ref_charged_atoms.append(idx)
    
    for (k,v) in mapping.items():
        if k in ref_charged_atoms or v in mol_charged_atoms:
            mol.GetAtomWithIdx(v).SetFormalCharge(ref.GetAtomWithIdx(k).GetFormalCharge())

    return mol

def reassign_bond_orders(mol, ref, mapping):
    #TODO this could be optimized
    for i in mapping.keys():
        for ref_bond in ref.GetAtomWithIdx(i).GetBonds():
            j = ref_bond.GetOtherAtomIdx(i)
            if j in mapping.keys():
                mol_bond = mol.GetBondBetweenAtoms(mapping[i], mapping[j])
                mol_bond.SetBondType(ref_bond.GetBondType())
    return mol

def h_coord_from_dipeptide(pdb1, pdb2):
    mol = Chem.MolFromPDBBlock(pdb1+pdb2)
    if mol is None:
        print(pdb1)
        print(pdb2)
        raise RuntimeError
    mol_h = Chem.AddHs(mol, addCoords=True)
    ps = Chem.SmilesParserParams()
    ps.removeHs = False
    template = Chem.MolFromSmiles('C(=O)C([H])N([H])C(=O)C([H])N',ps)
    h_idx = 5
    atom_map = mapping_by_mcs(template, mol_h)
    
    return mol_h.GetConformer().GetAtomPosition(atom_map[h_idx])

def h_coord_random_n_terminal(mol, debug=False):
    mol_no_h = Chem.RemoveHs(mol)
    for atom in mol_no_h.GetAtoms():
        if atom.GetProp("atom_name") == "N":
            bb_n_idx = atom.GetIdx()
    mol_h = Chem.AddHs(mol_no_h, addCoords=True)
    #positions = mol_h.GetConformer().GetPositions()
    bb_n = mol_h.GetAtomWithIdx(bb_n_idx)
    for neighbor in bb_n.GetNeighbors():
        if neighbor.GetAtomicNum() == 1:
            return mol_h.GetConformer().GetAtomPosition(neighbor.GetIdx())


def _snap_to_int(value, tolerance=0.12):
    for inc in [-1, 0, 1]:
        if abs(value - int(value) - inc) <= tolerance:
            return int(value) + inc
    return None

def divide_int_gracefully(integer, weights, allow_equal_weights_to_differ=False):
    for weight in weights:
        if type(weight) not in [int, float] or weight < 0:
            raise ValueError("weights must be numeric and non-negative")
    if type(integer) != int:
        raise ValueError("integer must be integer")
    inv_total_weight = 1.0 / sum(weights)
    shares = [w * inv_total_weight for w in weights] # normalize
    result = [_snap_to_int(integer * s, tolerance=0.5) for s in shares]
    surplus = integer - sum(result)
    if surplus == 0:
        return result
    data = [(i, w) for (i, w) in enumerate(weights)]
    data = sorted(data, key=lambda x: x[1], reverse=True)
    idxs = [i for (i, _) in data] 
    if allow_equal_weights_to_differ:
        groups = [1 for _ in weights]
    else:
        groups = []
        last_weight = None
        for i in idxs:
            if weights[i] == last_weight:
                groups[-1] += 1
            else:
                groups.append(1)
            last_weight = weights[i]

    # iterate over all possible combinations of groups
    # this is potentially very slow
    nr_groups = len(groups)
    for j in range(1, 2**nr_groups):
        n_changes = 0
        combo = []
        for grpidx in range(nr_groups):
            is_changed = bool(j & 2**grpidx) 
            combo.append(is_changed)
            n_changes += is_changed * groups[grpidx]
        if n_changes == abs(surplus):
            break

    # add or subtract 1 to distribute surplus
    increment = surplus / abs(surplus)
    index = 0
    for i, is_changed in enumerate(combo):
        if is_changed:
            for j in range(groups[i]):
                result[idxs[index]] += increment
                index += 1

    return result

def rectify_charges(q_list, net_charge=None, decimals=3):
    """make them 3 decimals and sum to an integer"""

    fstr = "%%.%df" % decimals
    charges_dec = [float(fstr % q) for q in q_list]

    if net_charge is None:
        net_charge = _snap_to_int(sum(charges_dec), tolerance=0.15)
        if net_charge is None:
            msg = "net charge could not be predicted from input q_list. (residual is beyond tolerance) "
            msg = "Please set the net_charge argument directly"
            raise RuntimeError(msg)
    elif type(net_charge) != int:
        raise TypeError("net charge must be an integer")

    surplus = net_charge - sum(charges_dec)
    surplus_int = _snap_to_int(10**decimals * surplus)

    if surplus_int == 0:
        return charges_dec

    weights = [abs(q) for q in q_list]
    surplus_int_splits = divide_int_gracefully(surplus_int, weights)
    for i, increment in enumerate(surplus_int_splits):
        charges_dec[i] += 10**-decimals * increment

    return charges_dec

class LinkedRDKitChorizo:
    """
    This is a class used to represent linked residues and associate them with RDKit and other information
    and there are most likely many more details that should be included in this little blurb.

    Attributes
    ----------
    cterm_pad_smiles: string
    nterm_pad_smiles: string
    backbone_smarts: string
    backbone: rdkit.Chem.rdchem.Mol
    backboneh: rdkit.Chem.rdchem.Mol
    nterm_pad_backbone_smarts_idxs: tuple of ints
    cterm_pad_backbone_smarts_idxs: tuple of ints
    rxn_cterm_pad: RDKit ChemicalReaction
    rxn_nterm_pad: RDKit ChemicalReaction

    TODO: Organize the following in a way that makes sense
    residues: dict (string -> ChorizoResidue) #TODO: figure out exact SciPy standard for dictionary key/value notation
    termini: dict (string (representing residue id) -> string (representing what we want the capping to look like))
    deleted_residues: list (string) residue ids to be deleted
    mutate_res_dict: dict (string (representing starting residue id) -> string (representing the desired mutated id))
    res_templates: dict (string -> rdkit.Chem.rdchem.Mol)
    ambiguous:
    disulfide_bridges:
    suggested_mutations:
    """

    cterm_pad_smiles = "CN"
    nterm_pad_smiles = "CC=O"
    backbone_smarts = "[C:1](=[O:2])[C:3][N:4]" # TODO make sure it matches res only once
    backbone = Chem.MolFromSmarts(backbone_smarts)
    backboneh = Chem.MolFromSmarts("[C:1](=[O:2])[C:3][N:4][#1]")
    nterm_pad_backbone_smarts_idxs = (0, 2, 1)
    cterm_pad_backbone_smarts_idxs = (2, 3)
    rxn_cterm_pad = rdChemReactions.ReactionFromSmarts(f"[N:5][C:6].{backbone_smarts}>>[C:6][N:5]{backbone_smarts}")
    rxn_nterm_pad = rdChemReactions.ReactionFromSmarts(f"[C:5][C:6]=[O:7].{backbone_smarts}>>{backbone_smarts}[C:6](=[O:7])[C:5]")


    def __init__(self, pdb_string, params=chorizo_params, mutate_res_dict=None, termini=None, deleted_residues=None, allow_bad_res=False):
        suggested_mutations = {}
        
        # Generates the residue representations based purely on the information in the pdb file to begin with.
        self.residues = self._pdb_to_resblocks(pdb_string)
        res_list = self.residues.keys()
        
        # User-specified modifications to the residues, capping of terminal residues and marking residues 
        # as deleted.
        self.termini = self._check_termini(termini, res_list)
        if deleted_residues is None:
            deleted_residues = ()
        self._check_del_res(deleted_residues, self.residues)
        self.deleted_residues = deleted_residues
        
        # User-specified mutations to the residues that we're tracking
        self.mutate_res_dict = mutate_res_dict
        if mutate_res_dict is not None:
            self._rename_residues(mutate_res_dict)

        # Loads user specified parameters
        self.res_templates, self.ambiguous = self._load_params(params)
        
        # Uses termini and user parameters to 
        ambiguous_chosen = self.parameterize_residues(self.termini, self.ambiguous)
        suggested_mutations.update(ambiguous_chosen)

        removed_residues = self.getIgnoredResidues()
        if len(removed_residues) > 0 and not allow_bad_res:
            for res in removed_residues:
                suggested_mutations[res] = res
            print("The following mutations are suggested. For HIS, mutate to HID, HIE, or HIP.")
            print(json.dumps(suggested_mutations, indent=2))
            msg = "The following residues could not be processed:" + pathlib.os.linesep
            msg += self.print_residues_by_resname(removed_residues)
            raise RuntimeError(msg)

        self.disulfide_bridges = self._find_disulfide_bridges()
        for cys_1, cys_2 in self.disulfide_bridges:
            chain_1, resname_1, resnum_1 = cys_1.split(":")
            chain_2, resname_2, resnum_2 = cys_2.split(":")
            if resname_1 != "CYX" or resname_2 != "CYX":
                print(f"Likely disulfide bridge between {cys_1} and {cys_2}")
            if resname_1 != "CYX":
                cyx_1 = f"{chain_1}:CYX:{resnum_1}"
                suggested_mutations[cys_1] = cyx_1
                if (cys_1 not in mutate_res_dict) and ((cys_2 not in mutate_res_dict) or resname_2 == "CYX"):
                    self._rename_residues({cys_1: cyx_1})
                    resmol = self.build_resmol(cyx_1, "CYX")
                    if resmol is not None:
                        self.residues[cyx_1].rdkit_mol = resmol
                    else:
                        self.residues[cyx_1].ignore_residue = True
            if resname_2 != "CYX":
                cyx_2 = f"{chain_2}:CYX:{resnum_2}"
                suggested_mutations[cys_2] = cyx_2
                if (cys_2 not in mutate_res_dict) and ((cys_1 not in mutate_res_dict) or resname_1 == "CYX"):
                    self._rename_residues({cys_2: cyx_2})
                    resmol = self.build_resmol(cyx_2, "CYX")
                    if resmol is not None:
                        self.residues[cyx_2].rdkit_mol = resmol
                    else:
                        self.residues[cyx_2].ignore_residue = True

        """to_remove = []
        for res_id in self.getIgnoredResidues():
            i = self.res_list.index(res_id)
            to_remove.append(i)
        for i in sorted(to_remove, reverse=True):
            self.res_list.pop(i) """
        self.suggested_mutations = suggested_mutations
        return

    def _find_disulfide_bridges(self):
        cys_list = {}
        cutoff = 2.5 # angstrom
        bridges = []
        for res in self.residues:
            if self.residues[res].ignore_residue: continue
            resname = res.split(":")[1]
            if resname in ["CYS", "CYX", "CYM"]: # TODO move "protected resnames" next to residue params they are associated with
                resmol = self.residues[res].rdkit_mol
                molxyz = resmol.GetConformer().GetPositions()
                s_xyz = None
                for atom in resmol.GetAtoms():
                    if atom.GetAtomicNum() == 16:
                        s_xyz = molxyz[atom.GetIdx()]
                for cys, other_s_xyz in cys_list.items():
                    v = s_xyz - other_s_xyz
                    dist = np.sqrt(np.dot(v, v))
                    if dist < cutoff:
                        bridges.append((res, cys))
                cys_list[res] = s_xyz
        return bridges

    @staticmethod
    def _check_del_res(query_res, residues):
        missing = set()
        for res in query_res:
            if res not in residues:
                missing.add(res)
            else:
                residues[res].user_deleted = True
        if len(missing) > 0:
            msg = "deleted_residues not found: " + " ".join(missing)
            raise ValueError(msg)

    @staticmethod
    def _check_termini(termini, res_list):
        allowed_c = ("cterm", "c-term", "c")
        allowed_n = ("nterm", "n-term", "n")
        output = {}
        if termini is None:
            return output
        for (resn, value) in termini.items():
            if resn not in res_list:
                raise ValueError("%s in termini not found" % resn)
            output[resn] = []
            if value.lower() in allowed_c:
                output[resn] = "C"
            elif value.lower() in allowed_n:
                output[resn] = "N"
            else:
                raise ValueError("termini value was %s, expected %s or %s" % (value, allowed_c, allowed_n))
        return output

    def get_padded_mol(self, resn):
        #TODO disulfides, ACE, NME
        # TODO double check next/previous res logic for "blunt" ending
        def _join(mol, pad_mol, pad_smarts_mol, rxn, is_res_atom, mapidx, adjacent_mol=None, pad_smarts_idxs=None):
            pad_matches = adjacent_mol.GetSubstructMatches(pad_smarts_mol)
            if len(pad_matches) != 1:
                raise RuntimeError("expected 1 match but got %d" % (len(pad_matches)))
            conformer = Chem.Conformer(pad_mol.GetNumAtoms())
            pad_mol.AddConformer(conformer)
            if adjacent_mol is not None:
                for index, smarts_index in enumerate(pad_smarts_idxs):
                    adjacent_mol_index = pad_matches[0][smarts_index]
                    pos = adjacent_mol.GetConformer().GetAtomPosition(adjacent_mol_index)
                    pad_mol.GetConformer().SetAtomPosition(index, pos)
            products, index_map = react_and_map((pad_mol, mol), rxn)
            if len(products) != 1:
                raise RuntimeError("expected 1 reaction product but got %d" % (len(products)))
            mol = products[0][0]
            index_map["reactant_idx"] = index_map["reactant_idx"][0][0]
            index_map["atom_idx"] = index_map["atom_idx"][0][0]
            Chem.SanitizeMol(mol)
            new_is_res_atom = []
            new_mapidx = {}
            for atom in mol.GetAtoms():
                index = atom.GetIdx()
                reactant_idx = index_map["reactant_idx"][index]
                if  reactant_idx == 0:
                    new_is_res_atom.append(False)
                elif reactant_idx == 1: # mol is 2nd reactant (0-index)
                    atom_idx = index_map["atom_idx"][index] 
                    new_is_res_atom.append(is_res_atom[atom_idx])
                    if atom_idx in mapidx:
                        new_mapidx[index] = mapidx[atom_idx]
                else:
                    raise RuntimeError("we have only two reactants, got %d ?" % reactant_idx)
            return mol, new_is_res_atom, new_mapidx

        mol = Chem.Mol(self.residues[resn].rdkit_mol)
        is_res_atom = [True for atom in mol.GetAtoms()]
        mapidx = {atom.GetIdx(): atom.GetIdx() for atom in mol.GetAtoms()}
        if self.residues[resn].previous_id is not None and self.residues[self.residues[resn].previous_id].rdkit_mol is not None:
            prev_resn = self.residues[resn].previous_id
            prev_mol = self.residues[prev_resn].rdkit_mol
            nterm_pad = Chem.MolFromSmiles(self.nterm_pad_smiles)
            mol, is_res_atom, mapidx = _join(
                    mol,
                    nterm_pad,
                    self.backbone,
                    self.rxn_nterm_pad,
                    is_res_atom,
                    mapidx,
                    prev_mol,
                    self.nterm_pad_backbone_smarts_idxs)
            
        if self.residues[resn].next_id is not None and self.residues[self.residues[resn].next_id].rdkit_mol is not None:
            next_resn = self.residues[resn].next_id
            next_mol = self.residues[next_resn].rdkit_mol
            cterm_pad = Chem.MolFromSmiles(self.cterm_pad_smiles)
            mol, is_res_atom, mapidx = _join(
                    mol,
                    cterm_pad,
                    self.backbone,
                    self.rxn_cterm_pad,
                    is_res_atom,
                    mapidx,
                    next_mol, 
                    self.cterm_pad_backbone_smarts_idxs)
        n_atoms_before_addhs = mol.GetNumAtoms()
        mol = Chem.AddHs(mol)
        is_res_atom.extend([False] * (mol.GetNumAtoms() - n_atoms_before_addhs))
        return mol, is_res_atom, mapidx

    def res_to_molsetup(self, res, mk_prep, is_protein_sidechain=False, cut_at_calpha=False):
        padded_mol, is_res_atom, mapidx = self.get_padded_mol(res)
        if is_protein_sidechain:
            bb_matches = padded_mol.GetSubstructMatches(self.backboneh)
            if len(bb_matches) != 1:
                raise RuntimeError("expected 1 backbone match, got %d" % (len(bb_matches)))
            c_alpha = bb_matches[0][2]
        else:
            c_alpha = None
        molsetups = mk_prep.prepare(padded_mol, root_atom_index=c_alpha)
        if len(molsetups) > 1:
            raise NotImplementedError("multiple molsetups not yet implemented for flexres")
        molsetup = molsetups[0]
        molsetup.is_sidechain = is_protein_sidechain
        ignored_in_molsetup = []
        for atom_index in molsetup.atom_ignore:
            if atom_index < len(is_res_atom):
                is_res = is_res_atom[atom_index] # Hs from Chem.AddHs beyond length of is_res_atom
            else:
                is_res = False
            ignore = not is_res
            ignore |= is_protein_sidechain and cut_at_calpha and (
                    (atom_index != c_alpha) and (atom_index in bb_matches[0]))
            molsetup.atom_ignore[atom_index] |= ignore
            if ignore and is_res:
                ignored_in_molsetup.append(mapidx[atom_index])
        # rectify charges to sum to integer (because of padding)
        net_charge = sum([atom.GetFormalCharge() for atom in self.residues[res].rdkit_mol.GetAtoms()])
        if mk_prep.charge_model == "zero":
            net_charge = 0
        not_ignored_idxs = []
        charges = []
        for i, q in molsetup.charge.items(): # charge is ordered dict
            if i in mapidx:
                charges.append(q)
                not_ignored_idxs.append(i) 
        charges = rectify_charges(charges, net_charge, decimals=3)
        for i, j in enumerate(not_ignored_idxs):
            molsetup.charge[j] = charges[i]
        return molsetup, mapidx, ignored_in_molsetup


    def flexibilize_protein_sidechain(self, res, mk_prep, cut_at_calpha=False):
        molsetup, mapidx, ignored_in_molsetup = self.res_to_molsetup(res, mk_prep,
                                                                  is_protein_sidechain=True,
                                                                  cut_at_calpha=cut_at_calpha)
        self.residues[res].molsetup = molsetup
        self.residues[res].molsetup_mapidx = mapidx
        self.residues[res].molsetup_ignored = ignored_in_molsetup
        self.residues[res].is_movable = True
        return

    def add_molsetup_inflexible(self, res, mk_prep, cut_at_calpha=False):
        molsetup, mapidx, ignored_in_molsetup = self.res_to_molsetup(res, mk_prep,
                                                                  is_protein_sidechain=False,
                                                                  cut_at_calpha=cut_at_calpha)  
        self.residues[res].molsetup = molsetup
        self.residues[res].molsetup_mapidx = mapidx
        self.residues[res].molsetup_ignored = ignored_in_molsetup
        self.residues[res].is_movable = False
        return

    @staticmethod
    def print_residues_by_resname(removed_residues):
        by_resname = dict()
        for res_id in removed_residues:
            chain, resn, resi = res_id.split(":")
            by_resname.setdefault(resn, [])
            by_resname[resn].append(f"{chain}:{resi}")
        string = ""
        for resname, removed_res in by_resname.items():
            string += f"Resname: {resname}:" + pathlib.os.linesep
            string += " ".join(removed_res) + pathlib.os.linesep
        return string

    @staticmethod 
    def _load_params(params):
        undesired_props = ("bonds", "//", "bond_cut_atoms", "smiles")
        
        ps = Chem.SmilesParserParams()
        ps.removeHs = False

        res_templates = {}
        for resn in params:
            if resn == "ambiguous": continue
            resmol = Chem.MolFromSmiles(params[resn]['smiles'], ps)
            # Using user provided parameters to assign properties to atoms.
            for idx, atom in enumerate(resmol.GetAtoms()):
                for propname in params[resn]:
                    if propname in undesired_props:
                        continue
                    value = params[resn][propname][idx]
                    if value is None:
                        continue
                    if type(value) == bool:
                        atom.SetBoolProp(propname, value)
                    elif type(value) == float:
                        atom.SetDoubleProp(propname, value)
                    elif type(value) == str:
                        atom.SetProp(propname, value)
                    else:
                        raise RuntimeError("property type:", type(value), value, "propname:", propname, "resn:", resn)
            res_templates[resn] = resmol
        ambiguous = params["ambiguous"]
        return res_templates, ambiguous
    
    @staticmethod
    def _pdb_to_resblocks(pdb_string):
        #TODO detect (and test distance) chain breaks
        #TODO cyclic peptides nex res == None ?!
        residues = {}
        # Tracking the key and the value for the current dictionary pair being read in
        current_res_id = None # the residue id we are tracking
        current_res = None    # the ChorizoResidue object we are tracking
        for line in pdb_string.splitlines(True):
            if line.startswith('TER') and current_res is not None:
                current_res.next_id = None
                residues[current_res_id] = current_res
                current_res = None
                current_res_id = None
            if line.startswith('ATOM') or line.startswith('HETATM'):
                # Generating dictionary key
                resname = line[17:20].strip()
                resid = int(line[22:26].strip())
                chainid = line[21].strip()
                full_res_id = ':'.join([chainid, resname, str(resid)])

                if full_res_id == current_res_id:
                    current_res.pdb_text += line
                else:
                    if current_res_id is not None:
                        last_resid = int(current_res_id.split(":")[2])
                        if resid - last_resid < 2:
                            current_res.next_id = full_res_id
                        else: # chain break
                            current_res.next_id = None
                    
                    # Updates tracking to the new key-value pair we're dealing with
                    current_res = ChorizoResidue(full_res_id, line)
                    if current_res_id is not None and (resid - int(current_res_id.split(":")[2])) < 2:
                        current_res.previous_id = current_res_id
                    else:
                        current_res.previous_id = None
                    current_res_id = full_res_id
                    residues[current_res_id] = current_res
        if current_res is not None:
            current_res.next_id = None
            residues[current_res_id] = current_res
        return residues
    
    def _rename_residues(self, mutate_dict):
        residue_order = list(self.residues.keys())
        for res in mutate_dict:
            old_resn = res.split(':')[1]
            new_resn = mutate_dict[res].split(':')[1]
            # Adds the previous residue as a new key in the dictionary
            self.residues[mutate_dict[res]] = self.residues.pop(res)
            # modifies the residue id so it is consistent with the new name we are giving the residue
            # TODO: Check if we want to do this or leave the original id in there
            self.residues[mutate_dict[res]].residue_id = mutate_dict[res]

            # updates dictionary ids
            previous_res = self.residues[mutate_dict[res]].previous_id
            if previous_res:
                self.residues[previous_res].next_id = mutate_dict[res]
            next_res = self.residues[mutate_dict[res]].next_id
            if next_res:
                self.residues[next_res].previous_id = mutate_dict[res]

            # tracks locations for reordering
            i = residue_order.index(res)
            residue_order[i] = mutate_dict[res]
        
        # clears and recreates self.residues in the desired order. Fairly inefficient
        # but we should only be doing this once.
        for residue in residue_order:
            value = self.residues.pop(residue)
            self.residues[residue] = value

    @staticmethod
    def add_termini(resn, res, termini, residues):
        # resn, res, termini, self.residues -> input
        next_res = residues[res].next_id
        prev_res = residues[res].previous_id
        if termini.get(res, None) == "C":
            if (next_res is not None) and (not residues[next_res].user_deleted):
                raise ValueError("Trying to C-term {res} but {next_res=} exists")
            resn = 'C' + resn
        elif termini.get(res, None) == "N":
            if (prev_res is not None) and (not residues[prev_res].user_deleted):
                raise ValueError("Trying to N-term {res} but {prev_res=} exists")
            resn = 'N' + resn
        elif termini.get(res, None) is None:
            resn = resn # wow, such assignment, very code
        else:
            # TODO verify sanity of termini earlier
            raise ValueError("termini must be either 'C' or 'N', not %s" % termini.get(res, None))
        return resn

    def parameterize_residues(self, termini, ambiguous):
        ambiguous_chosen = {}
        for res in self.residues:
            # skip deleted resides
            if self.residues[res].user_deleted:
                continue
            
            # if we can't generate an RDKit Mol for the residue, mark it as ignored and skip
            pdbmol = Chem.MolFromPDBBlock(self.residues[res].pdb_text, removeHs=False)
            if pdbmol is None:
                self.residues[res].ignore_residue = True
                continue
            
            # Check if parameters are available for a residue
            chain, resn, resnum = res.split(':')
            if (resn not in self.res_templates) and (resn not in ambiguous):
                #self.residues.pop(res)
                self.residues[res].ignore_residue = True
                continue

            if resn in ambiguous:
                possible_resn = ambiguous[resn]
            else:
                possible_resn = [resn]

            #if resn == "HIS": print("HIS ambiguous:", possible_resn)

            lowest_nr_missing = 9999999
            # loops through possible resmol generated for the residue based on number of atoms missing
            for resn in possible_resn:

                resn = self.add_termini(resn, res, termini, self.residues) # prefix C or N if applicable

                # TODO add to preprocessing to save time
                # Create mol object and map between the pdb and residue template
                resmol = Chem.Mol(self.res_templates[resn])
                n_atoms = len(resmol.GetAtoms())
                atom_map = mapping_by_mcs(resmol, pdbmol)
                nr_missing = n_atoms - len(atom_map)
                if nr_missing < lowest_nr_missing:
                    best_resmol = resmol
                    lowest_nr_missing = nr_missing
                    best_n_atoms = n_atoms
                    best_resn = resn
            resmol = best_resmol
            n_atoms = best_n_atoms
            resn = best_resn
            if len(possible_resn) > 1:
                print("%9s" % res, "-->", resn, "...out of", possible_resn)
                ambiguous_chosen[res] = f"{chain}:{resn}:{resnum}"

            resmol = self.build_resmol(res, resn)
            if resmol is None:
                self.residues[res].ignore_residue = True
            else:
                self.residues[res].rdkit_mol = resmol
        return ambiguous_chosen

    def build_resmol(self, res, resn):
        # Transfer coordinates and info for any matched atoms
        #TODO time these functions
        #TODO maybe embed in preprocessing depending on time
        #EmbedMolecule(resmol)

        resmol = Chem.Mol(self.res_templates[resn])
        pdbmol = Chem.MolFromPDBBlock(self.residues[res].pdb_text, removeHs=False)

        atom_map = mapping_by_mcs(resmol, pdbmol)
        #Chem.rdDepictor.Compute2DCoords(resmol)
        resmol.AddConformer(Chem.Conformer(resmol.GetNumAtoms()))

        resmol.GetConformer().Set3D(True)
        for idx, pdb_idx in atom_map.items():
            pdb_atom = pdbmol.GetAtomWithIdx(pdb_idx)
            pdb_coord = pdbmol.GetConformer().GetAtomPosition(pdb_idx)
            resmol.GetConformer().SetAtomPosition(idx, pdb_coord)

            resinfo = pdb_atom.GetPDBResidueInfo()
            resmol.GetAtomWithIdx(idx).SetDoubleProp('occupancy', resinfo.GetOccupancy())
            resmol.GetAtomWithIdx(idx).SetDoubleProp('temp_factor', resinfo.GetTempFactor())

        missing_atoms = {resmol.GetAtomWithIdx(i).GetProp('atom_name'):i for i in range(resmol.GetNumAtoms()) if i not in atom_map.keys()}

        # Handle case of missing backbone amide H
        if 'H' in missing_atoms:
            prev_res = self.residues[res].previous_id
            if prev_res is not None:
                h_pos = h_coord_from_dipeptide(self.residues[res].pdb_text, 
                                                self.residues[prev_res].pdb_text)
            else:
                h_pos = h_coord_random_n_terminal(resmol)
            resmol.GetConformer().SetAtomPosition(missing_atoms['H'], h_pos)
            resmol.GetAtomWithIdx(missing_atoms['H']).SetBoolProp('computed', True)
            missing_atoms.pop('H')
            
        
        missing_atom_elements = set([atom[0] for atom in missing_atoms.keys()])
        if len(missing_atom_elements) > 0:
            
            resmol_h = Chem.RemoveHs(resmol)
            resmol_h = Chem.AddHs(resmol_h, addCoords=True)
            h_map = mapping_by_mcs(resmol, resmol_h)
            for atom in list(missing_atoms.keys()):
                if atom.startswith('H'):
                    h_idx = missing_atoms[atom]
                    #TODO magic function from Diogo to add H atom
                    #TODO bring in parameterless H atom for chain breaks
                    resmol.GetConformer().SetAtomPosition(h_idx, resmol_h.GetConformer().GetAtomPosition(h_map[h_idx]))
                    resmol.GetAtomWithIdx(h_idx).SetBoolProp('computed', True)
                    missing_atoms.pop(atom)
        # TODO missing atoms from PDB? Extra PDB atoms OK currently?
        if len(missing_atoms) > 0:
            err = f'Could not add {res=} {missing_atoms=}'
            print(err)
            resmol = None
        return resmol


    def mk_parameterize_all_residues(self, mk_prep):
        # TODO disulfide bridges are hard-coded, generalize branching maybe
        for res in self.getAllValidResidues():
            """if self.residues[res].user_deleted or self.residues[res].ignore_residue:
                continue"""
            self.mk_parameterize_residue(self, res, mk_prep)
        return


    def mk_parameterize_residue(self, res, mk_prep): 
        molsetup, mapidx, ignored_in_molsetup = self.res_to_molsetup(res, mk_prep)
        resmol = self.residues[res].rdkit_mol
        for molsetup_idx, resmol_idx in mapidx.items(): # ignoring pseudo atoms, rdkit atom props no good for pseudos
            atom = resmol.GetAtomWithIdx(resmol_idx)
            atom.SetDoubleProp("q", molsetup.charge[molsetup_idx])
            atom.SetProp("atom_type", molsetup.atom_type[molsetup_idx])
            #atom.SetProp("ignore", molsetup.atom_ignore[molsetup_idx])
            for key, value_array in molsetup.atom_params.items():
                value = value_array[molsetup_idx]
                if type(value) == float:
                    atom.SetDoubleProp(key, value)
                elif type(value) == bool:
                    atom.SetBoolProp(key, value)
                else:
                    atom.SetProp(key, value)
        # consider deleting existing properties/parameters
        return
                

    def to_pdb(self, use_modified_coords=False, modified_coords_index=0):
        pdbout = ""
        atom_count = 0
        icode = ""
        pdb_line = "{:6s}{:5d} {:^4s} {:3s} {:1s}{:4d}{:1s}   {:8.3f}{:8.3f}{:8.3f}                       {:2s} "
        pdb_line += pathlib.os.linesep
        for res_id in self.residues:
            if self.residues[res_id].user_deleted or self.residues[res_id].ignore_residue:
                continue
            resmol = self.residues[res_id].rdkit_mol
            if use_modified_coords and self.residues[res_id].molsetup is not None:
                molsetup = self.residues[res_id].molsetup
                if len(molsetup.modified_atom_positions) <= modified_coords_index:
                    errmsg = "Requesting pose %d but only got %d in molsetup of %s" % (
                            modified_coords_index, len(molsetup.modified_atom_positions), res_id)
                    raise RuntimeError(errmsg)
                p = molsetup.modified_atom_positions[modified_coords_index]
                modified_positions = molsetup.get_conformer_with_modified_positions(p).GetPositions()
                positions = {}
                for i, j in self.residues[res_id].molsetup_mapidx.items():
                    positions[j] = modified_positions[i]
            else:
                positions = {i: xyz for (i, xyz) in enumerate(resmol.GetConformer().GetPositions())}

            chain, resname, resnum = res_id.split(":")
            resnum = int(resnum)

            for (i, atom) in enumerate(resmol.GetAtoms()):
                atom_count += 1
                props = atom.GetPropsAsDict()
                atom_name = props.get("atom_name", "")
                x, y, z = positions[i]
                element = mini_periodic_table[atom.GetAtomicNum()]
                pdbout += pdb_line.format("ATOM", atom_count, atom_name, resname, chain, resnum, icode, x, y, z, element)
        return pdbout

    def export_static_atom_params(self, ignore_atom_types=("H",)):
        atom_params = {}
        counter_atoms = 0
        coords = []
        for res_id in self.residues:
            if self.residues[res_id].user_deleted or self.residues[res_id].ignore_residue:
                continue
            resmol = self.residues[res_id].rdkit_mol
            for atom in resmol.GetAtoms():
                props = atom.GetPropsAsDict()
                if len(ignore_atom_types) > 0 and props["atom_type"] in ignore_atom_types:
                    continue
                if (self.residues[res_id].molsetup):
                    if atom.GetIdx() not in self.residues[res_id].molsetup_ignored:
                        continue
                for key, value in props.items():
                    if key.startswith("_"):
                        continue
                    atom_params.setdefault(key, [None]*counter_atoms) # add new "column"
                    atom_params[key].append(value)
                counter_atoms += 1
                for key in set(atom_params).difference(props): # <key> missing in <props>
                    atom_params[key].append(None) # fill in incomplete "row"
                coords.append(resmol.GetConformer().GetAtomPosition(atom.GetIdx()))
        if hasattr(self, "param_rename"): # e.g. "gasteiger" -> "q"
            for key, new_key in self.param_rename.items():
                atom_params[new_key] = atom_params.pop(key)
        return atom_params, coords
    
    # The following functions return filtered dictionaries of residues based on the value of residue flags.
    def getUserDeletedResidues(self):
        return {k: v for k, v in self.residues.items() if v.user_deleted == True}
        
    def getNonUserDeletedResidues(self):
        return {k: v for k, v in self.residues.items() if v.user_deleted == False}
        
    def getIgnoredResidues(self):
        return {k: v for k, v in self.residues.items() if v.ignore_residue == True}
        
    def getNotIgnoredResidues(self):
        return {k: v for k, v in self.residues.items() if v.ignore_residue == False}
    
    # TODO: rename this
    def getValidResidues(self):
        return {k: v for k, v in self.residues.items() if v.ignore_residue == False and v.user_deleted == False}


residues_rotamers = {"SER": [("C", "CA", "CB", "OG")],
                     "THR": [("C", "CA", "CB", "CG2")],
                     "CYS": [("C", "CA", "CB", "SG")],
                     "VAL": [("C", "CA", "CB", "CG1")],
                     "HIS": [("C", "CA", "CB", "CG"), 
                             ("CA", "CB", "CG", "CD2")],
                     "ASN": [("C", "CA", "CB", "CG"), 
                             ("CA", "CB", "CG", "ND2")],
                     "ASP": [("C", "CA", "CB", "CG"),
                             ("CA", "CB", "CG", "OD1")],
                     "ILE": [("C", "CA", "CB", "CG2"), 
                             ("CA", "CB", "CG2", "CD1")],
                     "LEU": [("C", "CA", "CB", "CG"), 
                             ("CA", "CB", "CG", "CD1")],
                     "PHE": [("C", "CA", "CB", "CG"), 
                             ("CA", "CB", "CG", "CD2")],
                     "TYR": [("C", "CA", "CB", "CG"), 
                             ("CA", "CB", "CG", "CD2")],
                     "TRP": [("C", "CA", "CB", "CG"), 
                             ("CA", "CB", "CG", "CD2")],
                     "GLU": [("C", "CA", "CB", "CG"), 
                             ("CA", "CB", "CG", "CD"), 
                             ("CB", "CG", "CD", "OE1")],
                     "GLN": [("C", "CA", "CB", "CG"), 
                             ("CA", "CB", "CG", "CD"), 
                             ("CB", "CG", "CD", "OE1")],
                     "MET": [("C", "CA", "CB", "CG"), 
                             ("CA", "CB", "CG", "SD"), 
                             ("CB", "CG", "SD", "CE")],
                     "ARG": [("C", "CA", "CB", "CG"), 
                             ("CA", "CB", "CG", "CD"), 
                             ("CB", "CG", "CD", "NE"),
                             ("CG", "CD", "NE", "CZ")],
                     "LYS": [("C", "CA", "CB", "CG"), 
                             ("CA", "CB", "CG", "CD"), 
                             ("CB", "CG", "CD", "CE"),
                             ("CG", "CD", "CE", "NZ")]}


rotamer_res_disambiguate = {}
for primary_res, specific_res_list in chorizo_params["ambiguous"].items():
    for specific_res in specific_res_list:
        rotamer_res_disambiguate[specific_res] = primary_res

def add_rotamers_to_chorizo_molsetups(rotamer_states_list, chorizo):
    no_resname_to_resname = {}
    for res_with_resname in chorizo.residues:
        chain, resname, resnum = res_with_resname.split(":")
        no_resname_key = f"{chain}:{resnum}"
        if no_resname_key in no_resname_to_resname:
            errmsg = "both %s and %s would be keyed by %s" % (
                res_with_resname, no_resname_to_resname[no_resname_key], no_resname_key)
            raise RuntimeError(errmsg)
        no_resname_to_resname[no_resname_key] = res_with_resname

    state_indices_list = []
    for state_dict in rotamer_states_list:
        state_indices = {}
        for res_no_resname, angles in state_dict.items():
            res_with_resname = no_resname_to_resname[res_no_resname]
            if not "molsetup" in chorizo.residues[res_with_resname]:
                raise RuntimeError("no molsetup for %s, can't add rotamers" % (res_with_resname))
            # next block is inneficient for large rotamer_states_list
            # refactored chorizos could help by having the following
            # data readily available
            resmol = chorizo.residues[res_with_resname]["resmol"]
            molsetup = chorizo.residues[res_with_resname]["molsetup"]
            mapidx = chorizo.residues[res_with_resname]["molsetup_mapidx"]
            mapidx_inv = {value: key for (key, value) in mapidx.items()}
            name_to_molsetup_idx = {}
            for atom in resmol.GetAtoms():
                props = atom.GetPropsAsDict()
                if "atom_name" in props:
                    atom_name = props["atom_name"]
                    name_to_molsetup_idx[atom_name] = mapidx_inv[atom.GetIdx()]

            resname = res_with_resname.split(":")[1]
            resname = rotamer_res_disambiguate.get(resname, resname)

            atom_names = residues_rotamers[resname]
            if len(atom_names) != len(angles):
                raise RuntimeError(
                    f"expected {len(atom_names)} angles for {resname}, got {len(angles)}")

            atom_idxs = []
            for names in atom_names:
                tmp = [name_to_molsetup_idx[name] for name in names]
                atom_idxs.append(tmp)

            state_indices[res_with_resname] = len(molsetup.rotamers)
            molsetup.add_rotamer(atom_idxs, np.radians(angles))

        state_indices_list.append(state_indices)

    return state_indices_list

class ChorizoResidue:
    """
    A class representing a single residue in the chain of chorizo residues

    Attributes
    ----------
    residue_id: string
        the residue id, most likely the key for this object if it is being 
        stored in a dictionary
    pdb_text: string
        the text from the pdb file associated with this residue
    previous_id: string or None
        the previous residue in this chain
    next_id: string or None
        the next residue in this chain
    
    rdkit_mol: rdkit.Chem.rdchem.Mol or None
        the rdkit Mol object generated from this residue
    molsetup: RDKitMoleculeSetup or None
        the RDKitMoleculeSetup object generated from this residue
    molsetup_mapidx: or None
        needs info
    molsetup_ignored: List[] or None
        needs info

    ignore_residue: bool
        marks residues that formerly were part of the removed_residues structure,
        put on residues that are ignored due to being incomplete or incorrect
    is_movable: bool
        marks residues that are flexible
    user_deleted: bool
        marks residues that the user indicated should be deleted

    additional_connections: List[ResidueAdditionalConnection]
        connections to additional residues along the chain (there's probably
        more specific information about the nature of these connections that
        could be added here)
    """
    def __init__(self, residue_id, pdb_text, previous_id=None, next_id=None):
        self.residue_id = residue_id
        self.pdb_text = pdb_text
        self.previous_id = previous_id
        self.next_id = next_id

        self.rdkit_mol = None
        self.molsetup = None
        self.molsetup_mapidx = None
        self.molsetup_ignored = None # Check about these data types/Do we want the default to be None or empty

        # flags
        self.ignore_residue = False
        self.is_movable = False
        self.user_deleted = False

        self.additional_connections = []

    def toJson(self):
        return json.dumps(self, default=lambda o: o.__dict__)
    
    def isValidResidue(self):
        """Returns true if the residue is not marked as deleted by a user and has not been marked as a residue to ignore"""
        return not self.ignore_residue and not self.user_deleted

# This could be a named tuple or a dataclass as it stands, but that is dependent on the amount of custom behavior
# we want to encode for these additional connections.
class ResidueAdditionalConnection:
    """
    Represents additional connections & bonds from a residue

    Attributes
    ----------
    connection_residue: string
        the id of the connected residue
    connection_atom: string
        the connected atom
    bond_order: string                    # SHOULD MAYBE NOT BE A STRING?
        the bond order of the connection
    """
    def __init__(self, residue, atom, bond_order):
        self.connection_residue = None
        self.connection_atom = None
        self.bond_order = None

    def toJson(self):
        return json.dumps(self, default=lambda o: o.__dict__)



#!/usr/bin/env python

import argparse
import json
import math
from os import linesep as os_linesep
import pathlib
import pickle
import sys


from meeko import PDBQTReceptor
from meeko import PDBQTMolecule
from meeko import RDKitMolCreate
from meeko import MoleculePreparation
from meeko import MoleculeSetup
from meeko import PDBQTWriterLegacy
from meeko import LinkedRDKitChorizo
from meeko import reactive_typer
from meeko import get_reactive_config
from meeko import gridbox
from rdkit import Chem

path_to_this_script = pathlib.Path(__file__).resolve()

# the following preservers RDKit Atom properties in the chorizo pickle
Chem.SetDefaultPickleProperties(Chem.PropertyPickleOptions.AtomProps) # |
#                                Chem.PropertyPickleOptions.PrivateProps)

def parse_residue_string(string):
    """
    Args:
        string (str): Residue identifier in the format 'chain:resname:resnum'.

    Returns:
        tuple: Residue identifier components: (chain, resname, resnum).
        bool: Whether the parsing was successful or not.
        str: An error message describing any encountered parsing errors.

    Example:
        >>> parse_residue_string("A:ALA:42")
        (('A', 'ALA', 42), True, '')
    """

    ok = True
    err = ""
    res_id = (None, None, None)
    if string.count(":") != 2:
        ok = False
        err += "Need exacly two ':' but found %d in '%s'" % (string.count(":"), string) + os_linesep
        err += "Example: 'A:HIE:42'" + os_linesep
        return res_id, ok, err
    chain, resname, resnum = string.split(":")
    if len(chain) not in (0, 1):
        ok = False
        err += "chain must be 0 or 1 char but it is '%s' (%d chars) in '%s'" % (chain, len(chain), string) + os_linesep
    if len(resname) > 3:
        ok = False
        err += "resname must be max 3 characters long, but is '%s' in '%s'" % (resname, string) + os_linesep
    try:
        resnum = int(resnum)
    except:
        ok = False
        err += "resnum could not be converted to integer, it was '%s' in '%s'" % (resnum, string) + os_linesep
    if ok:
        res_id = (chain, resname, resnum)
    return res_id, ok, err

def parse_residue_string_and_name(string):
    """
    Example:
        >>> parse_residue_string_and_name("A:HIE:42:CE1")
        ({"res_id": ('A', 'ALA', 42), "name": "CE1"}, True, '')
    """
    ok = True
    err = ""
    output = {"res_id": (None, None, None),
              "name": None}
    if string.count(":") != 3:
        ok = False
        err += "Need three ':' but found %d in %s" % (string.count(":"), string) + os_linesep
        err += "Example: 'A:HIE:42:CE1'" + os_linesep
        return output, ok, err

    resid_string, name = string.rsplit(":", 1)
    output["name"] = name
    res_id, ok_, err_ = parse_residue_string(resid_string)
    output["res_id"] = res_id
    ok &= ok_
    err += err_
    return output, ok, err

def parse_resname_and_name(string):
    """
    Example:
        >>> parse_resname_and_name("HIE:CE1")
        (("HIE", "CE1"), True, "")
    """
    ok = True
    err = ""
    output = (None, None)
    if string.count(":") != 1:
        ok = False
        err += "Expected one ':' but found %d in %s" % (string.count(":"), string) + os_linesep
        err += "Example: 'HIE:CE1'" + os_linesep
        return output, ok, err
    resname, name = string.split(":")
    return (resname, name), ok, err




class TalkativeParser(argparse.ArgumentParser):
    def error(self, message):
        """overload to print_help for every error"""
        self.print_help()
        this_script = path_to_this_script.name
        print('\n%s: error: %s' % (this_script, message), file=sys.stderr)
        sys.exit(2)

def check(success, error_msg):
    if not success:
        print("Error: " + error_msg, file=sys.stderr)
        sys.exit(2)

def get_args():
    parser = TalkativeParser()

    io_group = parser.add_argument_group("Input/Output")
    io_group.add_argument('--pdb', help="input can be PDBQT but charges and types will be reassigned")
    #parser.add_argument('--pdbqt', help="keeps existing charges and types")
    io_group.add_argument('-o', '--output_filename', required=True, help="adds _rigid/_flex with flexible residues. Always suffixes .pdbqt.")
    io_group.add_argument('-p', '--chorizo_pickle') 

    config_group = parser.add_argument_group("Receptor perception")
    config_group.add_argument('-n', '--mutate_residues',
                                  help="e.g. '{\"A:HIS:323\":\"A:HID:323\"}'")
    config_group.add_argument(      '--termini',
                                  help="e.g. '{\"A:GLY:350\":\"C-term\"}'")
    config_group.add_argument(      '--del_res', help="e.g. '[\"A:GLY:350\", \"B:ALA:17\"]'")
    config_group.add_argument(      '--chorizo_config', help="[.json]")
    config_group.add_argument(      '--mk_config', help="[.json]")
    config_group.add_argument(      '--allow_bad_res', action="store_true",
                                                 help="delete residues with missing atoms instead of raising error")
    config_group.add_argument('-f', '--flexres', action="append", default=[],
                        help="repeat flag for each residue, e.g: -f \" :LYS:42\" -f \"B:TYR:23\" and keep space for empty chain")

    box_group = parser.add_argument_group("Size and center of grid box")
    #box_group.add_argument('-b', '--gridbox_filename', help="set grid box size and center using a Vina configuration file")
    box_group.add_argument('--skip_gpf', help="do not write a GPF file for autogrid", action="store_true")
    box_group.add_argument('--box_size', help="size of grid box (x, y, z) in Angstrom", nargs=3, type=float)
    box_group.add_argument('--box_center', help="center of grid box (x, y, z) in Angstrom", nargs=3, type=float)
    box_group.add_argument('--box_center_on_reactive_res', help="project center of grid box along CA-CB bond 5 A away from CB", action="store_true")
    box_group.add_argument('--ligand', help="Reference ligand file path: .sdf, .mol, .mol2, .pdb, and .pdbqt files accepted")
    box_group.add_argument('--padding', help="padding around reference ligand [A]", type=float)


    #reactive_group = parser.add_argument_group("Reactive")
    #reactive_group.add_argument('-r', '--reactive_flexres', action="append", default=[],
    #                    help="same as --flexres but for reactive residues (max 8)")
    #reactive_group.add_argument('-g', '--reactive_name', action="append", default=[],
    #                    help="set name of reactive atom of a residue type, e.g: -g 'TRP:NE1'. Overridden by --reactive_name_specific")
    #reactive_group.add_argument('-s', '--reactive_name_specific', action="append", default=[],
    #                    help="set name of reactive atom for an individual residue, e.g: -s 'A:HIE:42:NE2'. Residue will be reactive.")

    #reactive_group.add_argument('--r_eq_12', default=1.8, type=float, help="r_eq for reactive atoms (1-2 interaction)")
    #reactive_group.add_argument('--eps_12', default=2.5, type=float, help="epsilon for reactive atoms (1-2 interaction)")
    #reactive_group.add_argument('--r_eq_13_scaling', default=0.5, type=float, help="r_eq scaling for 1-3 interaction across reactive atoms")
    #reactive_group.add_argument('--r_eq_14_scaling', default=0.5, type=float, help="r_eq scaling for 1-4 interaction across reactive atoms")
    args = parser.parse_args()

    #if (args.pdb is None) == (args.pdbqt is None):
        #msg = "need either --pdb or --pdbqt"
    if args.pdb is None:
        msg = "need --pdb"
        print("Command line error: " + msg, file=sys.stderr)
        sys.exit(2)
    if (args.box_center is not None) and args.box_center_on_reactive_res:
        msg = "can't use both --box_center and --box_center_on_reactive_res"
        print("Command line error: " + msg, file=sys.stderr)
        sys.exit(2)
    got_center = (args.box_center is not None) or args.box_center_on_reactive_res or (args.ligand is not None)
    if not args.skip_gpf:
        if not got_center:
            msg  = "missing center or size of grid box to write .gpf file for autogrid4" + os_linesep
            msg += "use --box_size and either --box_center or --box_center_on_reactive_res" + os_linesep
            msg += "or --ligand and --padding" + os_linesep
            msg += "Exactly one reactive residue required for --box_center_on_reactive_res" + os_linesep
            msg += "If a GPF file is not needed (e.g. docking with Vina scoring function) use option --skip_gpf"
            print("Command line error: " + msg, file=sys.stderr)
            sys.exit(2)
        if (args.box_size is None) and (args.padding is None):
            msg  = "grid box information is needed to dock with the AD4 scoring function." + os_linesep
            msg += "The grid box center and size will be used to write a GPF file for autogrid" + os_linesep
            msg += "If a GPF file is not needed (e.g. docking with Vina scoring function) use option --skip_gpf"
            print("Command line error: " + msg, file=sys.stderr)
            sys.exit(2)

    if (args.box_center is not None) + (args.ligand is not None) + args.box_center_on_reactive_res > 1:
        msg = "--box_center, --box_center_on_reactive_res, and --ligand are mutually exclusive options"
        print("Command line error: " + msg, file=sys.stderr)
        sys.exit(2)

    return args

args = get_args()

reactive_atom = {
    "SER":  "OG", "LYS":  "NZ", "TYR":  "OH", "CYS":  "SG", "HIE": "NE2",
    "HID": "ND1", "GLU": "OE2", "THR": "OG1", "MET":  "SD",
}
modified_resnames = set()
for react_name_str in args.reactive_name:
    (resname, name), ok, err = parse_resname_and_name(react_name_str)
    if ok:
        if resname in modified_resnames:
            print("Command line error: repeated resname %s passed to --reactive_resname" % resname + os_linesep, file=sys.stderr)
            sys.exit(2)
        modified_resnames.add(resname)
        reactive_atom[resname] = name
    else:
        print("Error in parsing --reactive_name argument" + os_linesep, file=sys.stderr)
        print(err, file=sys.stderr)
        sys.exit(2)

reactive_flexres = {}
all_ok = True
all_err = ""
for resid_string in args.reactive_flexres:
    res_id, ok, err = parse_residue_string(resid_string)
    if ok:
        resname = res_id[1]
        if resname in reactive_atom:
            reactive_flexres[res_id] = reactive_atom[resname]
        else:
            all_ok = False
            all_err += "no default reactive name for %s, " % resname
            all_err += "use --reactive_name or --reactive_name_specific" + os_linesep
    all_ok &= ok
    all_err += err

for string in args.reactive_name_specific:
    out, ok, err = parse_residue_string_and_name(string)
    if ok:
        # override name if res_id was also passed to --reactive_flexres
        reactive_flexres[out["res_id"]] = out["name"]
    all_ok &= ok
    all_err += err

if len(reactive_flexres) > 8:
    msg = "got %d reactive_flexres but maximum is 8." % (len(args.reactive_flexres))
    print("Command line error: " + msg, file=sys.stderr)
    sys.exit(2)

all_flexres = set()
for resid_string in args.flexres:
    res_id, ok, err = parse_residue_string(resid_string)
    all_ok &= ok
    all_err += err
    if ok:
        if res_id in reactive_flexres:
            all_err += "Command line error: can't determine if %s is reactive or not." % str(res_id) + os_linesep
            all_err += "Do not pass %s to --flexres if it is reactive." % str(res_id) + os_linesep
            all_ok = False
        all_flexres.add(res_id)

if all_ok:
    all_flexres = all_flexres.union(reactive_flexres)
else:
    print("Error:", file=sys.stderr)
    print(all_err, file=sys.stderr)
    sys.exit(2)

if len(all_flexres) > 0:
    print()
    print("Flexible residues:")
    print("chain resname resnum is_reactive reactive_atom")
    string = "%5s%8s%7d%12s%14s"
    for res_id in all_flexres:
        chain, resname, resnum = res_id
        is_react = res_id in reactive_flexres
        if is_react:
            react_atom = reactive_flexres[res_id]
        else:
            react_atom = ""
        print(string % (chain, resname, resnum, is_react, react_atom))
    print()

if len(reactive_flexres) != 1 and args.box_center_on_reactive_res:
    msg = "--box_center_on-reactive_res can be used only with one reactive" + os_linesep
    msg += "residue, but %d reactive residues are set" % len(reactive_flexres)
    print("Command line error:" + msg, file=sys.stderr)
    sys.exit(2)

if args.pdb is not None:
    mutate_res_dict = {}
    termini = {}
    del_res = []
    if args.chorizo_config is not None:
        with open(args.chorizo_config) as f:
            chorizo_config = json.load(f)
        mutate_res_dict.update(chorizo_config.get("mutate_res_dict", {}))
        termini.update(chorizo_config.get("termini", {}))
        del_res.extend(chorizo_config.get("del_res", []))
    # direct command line options override config
    if args.mutate_residues is not None:
        mutate_res_dict.update(json.loads(args.mutate_residues))
    if args.termini is not None:
        termini.update(json.loads(args.termini))
    if args.del_res is not None:
        del_res.update(json.loads(args.del_res))
    with open(args.pdb) as f:
        pdb_string = f.read()
    chorizo = LinkedRDKitChorizo(pdb_string, mutate_res_dict=mutate_res_dict, termini=termini, deleted_residues=del_res,
                                 allow_bad_res=args.allow_bad_res)
    if args.mk_config is not None:
        with open(args.mk_config) as f:
            mk_config = json.load(f)
        mk_prep = MoleculePreparation.from_config(mk_config)
        chorizo.mk_parameterize_all_residues(mk_prep)
    else:
        mk_prep = MoleculePreparation()


    for res_id in all_flexres:
        res = "%s:%s:%d" % res_id
        chorizo.flexibilize_protein_sidechain(res, mk_prep, cut_at_calpha=True)
    rigid_pdbqt, flex_pdbqt_dict = PDBQTWriterLegacy.write_from_linked_rdkit_chorizo(chorizo)
    if args.chorizo_pickle is not None:
        with open(args.chorizo_pickle, "wb") as f:
            pickle.dump(chorizo, f)

    suggested_config = {}
    if len(chorizo.suggested_mutations):
        suggested_config["mutate_res_dict"] = chorizo.suggested_mutations.copy()

    if len(chorizo.getIgnoredResidues()) > 0:
        print("Automatically deleted %d residues" % len(chorizo.removed_residues))
        print(json.dumps(chorizo.removed_residues, indent=4))
        suggested_config["del_res"] = chorizo.removed_residues.copy()

    #rigid_pdbqt, ok, err = PDBQTWriterLegacy.write_string_static_molsetup(molsetup)
    #ok, err = receptor.assign_types_charges()
    #check(ok, err)
#else:
#    receptor = PDBQTReceptor(args.pdbqt)
#    pdbqt, ok, err = receptor.write_pdbqt_string(flexres=all_flexres)
#    check(ok, err)

pdbqt = {
    "rigid": rigid_pdbqt,
    "flex": flex_pdbqt_dict,
}

any_lig_base_types = ["HD", "C", "A", "N", "NA", "OA", "F", "P", "SA",
                      "S", "Cl", "CL", "Br", "BR", "I", "Si", "B"]

outpath = pathlib.Path(args.output_filename)

written_files_log = {"filename": [], "description": []}

if len(all_flexres) == 0:
    box_center = args.box_center
    rigid_fn = str(outpath.with_suffix(".pdbqt"))
    flex_fn = None
else:
    all_flex_pdbqt = ""
    reactive_flexres_count = 0
    for res_id, flexres_pdbqt in pdbqt["flex"].items():
        res_id = tuple(res_id.split(":"))
        res_id = res_id[:2] + (int(res_id[2]),)
        if res_id in reactive_flexres:
            reactive_flexres_count += 1
            prefix_atype = "%d" % reactive_flexres_count
            resname = res_id[1]
            reactive_atom = reactive_flexres[res_id]
            flexres_pdbqt = PDBQTReceptor.make_flexres_reactive(flexres_pdbqt, reactive_atom, resname, prefix_atype)
        all_flex_pdbqt += flexres_pdbqt

    suffix = outpath.suffix
    if outpath.suffix == "":
        suffix = ".pdbqt"
    rigid_fn = str(outpath.with_suffix("")) + "_rigid" + suffix
    flex_fn = str(outpath.with_suffix("")) + "_flex" + suffix

    written_files_log["filename"].append(flex_fn)
    written_files_log["description"].append("flexible receptor input file")
    with open(flex_fn, "w") as f:
        f.write(all_flex_pdbqt)

written_files_log["filename"].append(rigid_fn)
written_files_log["description"].append("static (i.e., rigid) receptor input file")
with open(rigid_fn, "w") as f:
    f.write(pdbqt["rigid"])

if len(suggested_config):
    suggested_fn = str(outpath.with_suffix("")) + "_suggested-config.json"
    written_files_log["filename"].append(suggested_fn)
    written_files_log["description"].append("log of automated decisions for user inspection")
    with open(suggested_fn, "w") as f:
        json.dump(suggested_config, f)

# GPF for autogrid4
if not args.skip_gpf:
    if args.box_center is not None:
        box_center = args.box_center
        box_size = args.box_size
    elif args.box_center_on_reactive_res:
        # we have only one reactive residue and will set the box center
        # to be 5 Angstromg away from CB along the CA->CB vector
        idxs = receptor.atom_idxs_by_res[list(reactive_flexres.keys())[0]]
        ca = None
        cb = None
        for atom in receptor.atoms(idxs):
            if atom["name"] == "CA":
                ca = atom["xyz"]
            if atom["name"] == "CB":
                cb = atom["xyz"]
        if ca is None or cb is None:
            check(success=False, error_msg="could not find CA or CB in %s" % reactive_flexres[0])
        v = (cb - ca)
        v /= math.sqrt(v[0]**2 + v[1]**2 + v[2]**2) + 1e-8
        box_center = ca + 5 * v
        box_size = args.box_size
    elif args.ligand is not None:
        ft = pathlib.Path(args.ligand).suffix
        suppliers = {
            ".pdb": Chem.MolFromPDBFile,
            ".mol": Chem.MolFromMolFile,
            ".mol2": Chem.MolFromMol2File,
            ".sdf": Chem.SDMolSupplier,
            ".pdbqt": None
        }
        if ft not in suppliers.keys():
            check(success=False, error_msg=f"Given --ligand file type {ft} not readable!")
        elif ft != ".sdf" and ft != ".pdbqt":
            ligmol = suppliers[ft](args.ligand)
        elif ft == ".sdf":
            ligmol = suppliers[ft](args.ligand)[0]  # assume we only want first molecule in file
        else:  # .pdbqt
            ligmol = RDKitMolCreate.from_pdbqt_mol(PDBQTMolecule.from_file(args.ligand))[0]  # assume we only want first molecule in file
        
        box_center, box_size = gridbox.calc_box(ligmol.GetConformer().GetPositions(), args.padding)
    else:
        print("Error: No box center specified.", file=sys.stderr)
        sys.exit(2)

    # write .dat parameter file for B and Si
    ff_fn = pathlib.Path(rigid_fn).parents[0] / pathlib.Path("boron-silicon-atom_par.dat")
    written_files_log["filename"].append(str(ff_fn))
    written_files_log["description"].append("atomic parameters for B and Si (for autogrid)")
    with open(ff_fn, "w") as f:
        f.write(gridbox.boron_silicon_atompar)
    rec_types = ['HD', 'C', 'A', 'N', 'NA', 'OA', 'F', 'P', 'SA', 'S', 'Cl', 'Br', 'I', 'Mg', 'Ca', 'Mn', 'Fe', 'Zn']
    gpf_string, npts = gridbox.get_gpf_string(box_center, box_size, rigid_fn, rec_types, any_lig_base_types,
                                                ff_param_fname=ff_fn.name)
    # write GPF
    gpf_fn = pathlib.Path(rigid_fn).with_suffix(".gpf")
    written_files_log["filename"].append(str(gpf_fn))
    written_files_log["description"].append("autogrid input file")
    with open(gpf_fn, "w") as f:
        f.write(gpf_string)

    # write a PDB for the box
    box_fn = str(gpf_fn) + ".pdb"
    written_files_log["filename"].append(box_fn)
    written_files_log["description"].append("PDB file to visualize the grid box")
    with open(box_fn, "w") as f:
        f.write(gridbox.box_to_pdb_string(box_center, npts))

    # check all flexres are inside the box
    if len(reactive_flexres) > 0:
        any_outside = False
        for res_id, res in chorizo.residues.items():
            if not res.is_movable:
                continue
            for index, coord in res.molsetup.coord.items():
                if index in res.molsetup_ignored:
                    continue
                if gridbox.is_point_outside_box(atom["xyz"], box_center, npts):
                    print("WARNING: Flexible residue outside box." + os_linesep, file=sys.stderr)
                    print("WARNING: Strongly recommended to use a box that encompasses flexible residues." + os_linesep, file=sys.stderr)
                    break # only need to warn once

# configuration info for AutoDock-GPU reactive docking
if len(reactive_flexres) > 0:
    any_lig_reac_types = []
    for order in (1, 2, 3):
        for t in any_lig_base_types:
            any_lig_reac_types.append(reactive_typer.get_reactive_atype(t, order))

    rec_reac_types = []
    for line in all_flex_pdbqt.split(os_linesep):
        if line.startswith("ATOM") or line.startswith("HETATM"):
            atype = line[77:].strip()
            basetype, _ = reactive_typer.get_basetype_and_order(atype)
            if basetype is not None: # is None if not reactive
                rec_reac_types.append(line[77:].strip())

    derivtypes, modpairs, collisions = get_reactive_config(
                                    any_lig_reac_types,
                                    rec_reac_types,
                                    args.eps_12,
                                    args.r_eq_12,
                                    args.r_eq_13_scaling,
                                    args.r_eq_14_scaling)

    if len(collisions) > 0:
        collision_str = ""
        for t1, t2 in collisions:
            collision_str += "%3s %3s" % (t1, t2) + os_linesep
        collision_fn = str(outpath.with_suffix(".atype_collisions"))
        written_files_log["filename"].append(collision_fn)
        written_files_log["description"].append("type pairs (n=%d) that may lead to intra-molecular reactions" % len(collisions))
        with open(collision_fn, "w") as f:
            f.write(collision_str)

    # The maps block is to tell AutoDock-GPU the base types for the reactive types.
    # This could be done with -T/--derivtypes, but putting derivtypes and intnbp
    # lines in a single configuration file simplifies the command line call.
    map_block = ""
    map_prefix = pathlib.Path(rigid_fn).with_suffix("").name
    all_types = []
    for basetype, reactypes in derivtypes.items():
        all_types.append(basetype)
        map_block += "map %s.%s.map" % (map_prefix, basetype) + os_linesep
        for reactype in reactypes:
            all_types.append(reactype)
            map_block += "map %s.%s.map" % (map_prefix, basetype) + os_linesep
    config = "ligand_types " + " ".join(all_types) + os_linesep
    config += "fld %s.maps.fld" % map_prefix + os_linesep
    config += map_block

    # in modpairs (dict): types are keys, parameters are values
    # now we will write a configuration file with nbp keywords
    # that AD-GPU reads using the --import_dpf flag
    # nbp stands for "non-bonded potential" or "non-bonded pairwise"
    line = "intnbp_r_eps %8.6f %8.6f %3d %3d %4s %4s" + os_linesep
    nbp_count = 0
    for (t1, t2), param in modpairs.items():
        config += line % (param["r_eq"], param["eps"], param["n"], param["m"], t1, t2)
        nbp_count += 1
    config_fn = str(outpath.with_suffix(".reactive_config"))
    written_files_log["filename"].append(config_fn)
    written_files_log["description"].append("reactive parameters for AutoDock-GPU")
    with open(config_fn, "w") as f:
        f.write(config)
    print()
    print("For reactive docking, pass the configuration file to AutoDock-GPU:")
    print("    autodock_gpu -C 1 --import_dpf %s --flexres %s -L <ligand_filename>" % (config_fn, flex_fn))
    print()

print()
print("Files written:")
longest_fn = max([len(fn) for fn in written_files_log["filename"]])
line = "%%%ds <-- " % longest_fn + "%s"
for fn, desc in zip(written_files_log["filename"], written_files_log["description"]):
    print(line % (fn, desc))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#

__version__ = "0.6.0-alpha"

try:
    import openbabel
except ImportError:
    _has_openbabel = False
else:
    _has_openbabel = True

try:
    import prody
except ImportError:
    _has_prody = False
else:
    _has_prody = True

from .preparation import MoleculePreparation
from .molsetup import RDKitMoleculeSetup
from .molsetup import MoleculeSetup
from .molsetup import Restraint
from .molsetup import UniqAtomParams
from .utils import rdkitutils
from .utils import pdbutils
from .utils import geomutils
from .utils import utils
from .atomtyper import AtomTyper
from .receptor_pdbqt import PDBQTReceptor
from .linked_rdkit_chorizo import LinkedRDKitChorizo
from .linked_rdkit_chorizo import ChorizoResidue
from .linked_rdkit_chorizo import ResidueAdditionalConnection
from .linked_rdkit_chorizo import add_rotamers_to_chorizo_molsetups
from .molecule_pdbqt import PDBQTMolecule
from .rdkit_mol_create import RDKitMolCreate
from .reactive import reactive_typer
from .reactive import get_reactive_config
from .writer import PDBQTWriterLegacy
from . import analysis
from .writer import oids_block_from_setup
from .openff_xml_parser import parse_offxml
from .openff_xml_parser import load_openff
from .openff_xml_parser import get_openff_epsilon_sigma
from .hydrate import Hydrate

__all__ = ['MoleculePreparation', 'RDKitMoleculeSetup',
        'pdbutils', 'geomutils', 'rdkitutils', 'utils',
        'AtomTyper', 'PDBQTMolecule', 'PDBQTReceptor', 'analysis',
        'LinkedRDKitChorizo', 'ChorizoResidue', 'ResidueAdditionalConnection',
        'add_rotamers_to_chorizo_molsetups',
        'RDKitMolCreate',
        'PDBQTWriterLegacy',
        'reactive_typer',
        'get_reactive_config',
        'gridbox',
        'oids_block_from_setup',
        'parse_offxml',
        'Hydrate',
        'Restraint',
]

if _has_openbabel:
    from .molsetup import OBMoleculeSetup
    from .utils import obutils
    __all__.append("OBMoleculeSetup")
    __all__.append("obutils")

if _has_prody:
    from .covalentbuilder import CovalentBuilder
    __all__.append("CovalentBuilder")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko preparation
#

from inspect import signature
import json
import os
import pathlib
import sys
from collections import OrderedDict
import warnings

from rdkit import Chem

from .molsetup import OBMoleculeSetup
from .molsetup import RDKitMoleculeSetup
from .atomtyper import AtomTyper
from .espalomatyper import EspalomaTyper
from .bondtyper import BondTyperLegacy
from .hydrate import HydrateMoleculeLegacy
from .macrocycle import FlexMacrocycle
from .flexibility import FlexibilityBuilder
from .writer import PDBQTWriterLegacy
from .reactive import assign_reactive_types
from .openff_xml_parser import load_openff

pkg_dir = pathlib.Path(__file__).parents[0]
params_dir = pkg_dir / "data" / "params"
# the above is controversial, see
# https://stackoverflow.com/questions/6028000/how-to-read-a-static-file-from-inside-a-python-package

try:
    from openbabel import openbabel as ob
except ImportError:
    _has_openbabel = False
else:
    _has_openbabel = True

# DeprecationWarning is not displayed by default
warnings.filterwarnings("default", category=DeprecationWarning)

class MoleculePreparation:

    packaged_params = {}
    for path in params_dir.glob("*.json"): # e.g. data/params/ad4_types.json
        name = path.with_suffix("").name # e.g. "ad4_types"
        packaged_params[name] = path

    def __init__(self,
            merge_these_atom_types=("H",),
            hydrate=False,
            flexible_amides=False,
            rigid_macrocycles=False,
            min_ring_size=7,
            max_ring_size=33,
            keep_chorded_rings=False,
            keep_equivalent_rings=False,
            double_bond_penalty=50,
            rigidify_bonds_smarts=[],
            rigidify_bonds_indices=[],
            input_atom_params=None,
            load_atom_params="ad4_types",
            add_atom_types=(),
            input_offatom_params=None, 
            load_offatom_params=None,
            charge_model="gasteiger",
            dihedral_model=None,
            reactive_smarts=None,
            reactive_smarts_idx=None,
            add_index_map=False,
            remove_smiles=False,
        ):

        self.deprecated_setup_access = None
        self.merge_these_atom_types = merge_these_atom_types
        self.hydrate = hydrate
        self.flexible_amides = flexible_amides
        self.rigid_macrocycles = rigid_macrocycles
        self.min_ring_size = min_ring_size
        self.max_ring_size = max_ring_size
        self.keep_chorded_rings = keep_chorded_rings
        self.keep_equivalent_rings = keep_equivalent_rings
        self.double_bond_penalty = double_bond_penalty
        self.rigidify_bonds_smarts = rigidify_bonds_smarts
        self.rigidify_bonds_indices = rigidify_bonds_indices

        self.input_atom_params = input_atom_params
        self.load_atom_params = load_atom_params
        self.add_atom_types = add_atom_types

        self.atom_params = self.get_atom_params(
                input_atom_params,
                load_atom_params,
                add_atom_types,
                self.packaged_params)

        self.load_offatom_params = load_offatom_params
        
        allowed_charge_models = ["espaloma", "gasteiger", "zero"]
        if charge_model not in allowed_charge_models:
            raise ValueError("unrecognized charge_model: %s, allowed options are: %s" % (charge_model, allowed_charge_models))

        self.charge_model = charge_model
       
        allowed_dihedral_models = [None,'openff','espaloma']
        if dihedral_model in (None, 'espaloma'):
            dihedral_list = []
        elif dihedral_model == "openff":
            _, dihedral_list, _ = load_openff()
        else: 
            raise ValueError("unrecognized dihedral_model: %s, allowed options are: %s" % (dihedral_model, allowed_dihedral_models))
        
        self.dihedral_model = dihedral_model
        self.dihedral_params = dihedral_list

        if dihedral_model == 'espaloma' or charge_model == "espaloma":
            self.espaloma_model = EspalomaTyper()

        self.reactive_smarts = reactive_smarts
        self.reactive_smarts_idx = reactive_smarts_idx
        self.add_index_map = add_index_map
        self.remove_smiles = remove_smiles

        self._bond_typer = BondTyperLegacy()
        self._macrocycle_typer = FlexMacrocycle(
                self.min_ring_size, self.max_ring_size, self.double_bond_penalty)
        self._flex_builder = FlexibilityBuilder()
        self._water_builder = HydrateMoleculeLegacy()
        self._classes_setup = {Chem.rdchem.Mol: RDKitMoleculeSetup}

        self.offatom_params = {}

    @staticmethod
    def get_atom_params(input_atom_params, load_atom_params, add_atom_types, packaged_params):
        atom_params = {}
        if type(load_atom_params) == str:
            load_atom_params = [load_atom_params]
        elif load_atom_params is None:
            load_atom_params = ()
        for name in load_atom_params:
            filename = None
            if name == "openff-2.0.0" or name == "openff": # TODO allow multiple versions
                vdw_list, _, _ = load_openff() 
                d = {"openff-2.0.0": vdw_list}
            elif name in packaged_params:
                filename = packaged_params[name]
            elif name.endswith(".json"):
                filename = name
            else:
                msg =  "names passed to 'load_atom_params' need to suffixed with .json" + os.linesep
                msg += "or be the unsuffixed basename of a JSON file in %s." % str(params_dir) + os.linesep
                msg += "name was %s" % name 
                raise ValueError(msg)
            if filename is not None:
                with open(filename) as f:
                    d = json.load(f)
            overlapping_groups = set(atom_params).intersection(set(d))
            if len(overlapping_groups):
                msg = "overlapping parameter groups: %s" % str(overlapping_groups)
                raise ValueError(msg)
            atom_params.update(d)

        if input_atom_params is not None:
            d = json.loads(json.dumps(input_atom_params))
            overlapping_groups = set(atom_params).intersection(set(d))
            if len(overlapping_groups): # todo: remove duplicated code
                msg = "overlapping parameter groups: %s" % str(overlapping_groups)
                raise ValueError(msg)
            params_set_here = set()
            for group_key, rows in input_atom_params.items():
                for row in rows:
                    for param_name in row:
                        if param_name not in ["smarts", "comment", "IDX"]:
                            params_set_here.add(param_name)
            params_set_before = set()
            for group_key, rows in atom_params.items():
                for row in rows:
                    for param_name in row:
                        if param_name not in ["smarts", "comment", "IDX"]:
                            params_set_before.add(param_name)
            overlap = params_set_before.intersection(params_set_here)
            if len(overlap):
                msg = f"input_atom_params {overlap} also set by one or more of {load_atom_params}\n"
                msg += "consider setting load_atom_params=None"
                raise ValueError(f"input_atom_params {overlap} also set by one or more of {load_atom_params}")
                
            atom_params.update(d)

        if len(add_atom_types) > 0:
            group_keys = list(atom_params.keys())
            if len(group_keys) != 1:
                msg = "add_atom_types is usable only when there is one group of parameters"
                msg += ", but there are %d groups: %s" % (len(keys), str(keys))
                raise RuntimeError(msg)
            key = group_keys[0]
            atom_params[key].extend(add_atom_types)
               
        return atom_params

                ### else:
                ###     msg = "When atom_params is a list, it must consist of either:" + os.linesep
                ###     msg += "  - strings (to look up default JSON files in the data dir)" + os.linesep
                ###     msg += "  - or dictionaries (to be used directly). For example:" + os.linesep
                ###     msg += "                  {\"SMARTS\": \"[B]\", \"atype\": \"C\"}" + os.linesep
                ###     msg += "A total of %d types were found in atom_params:" % (len(types)) + os.linesep
                ###     msg += str(types) + os.linesep
                ###     raise ValueError(msg)
                ### 
                ### #for value in atom_params:
                    

        self.offatom_params = offatom_params
        self.charge_model = charge_model
        self.dihedral_params = dihedral_params


        if _has_openbabel:
            self._classes_setup[ob.OBMol] = OBMoleculeSetup
        if keep_chorded_rings and keep_equivalent_rings==False:
            warnings.warn("keep_equivalent_rings=False ignored because keep_chorded_rings=True", RuntimeWarning)
        if (reactive_smarts is None) != (reactive_smarts_idx is None):
            raise ValueError("reactive_smarts and reactive_smarts_idx require each other")

    @property
    def setup(self):
        msg = "MoleculePreparation.setup is deprecated in Meeko v0.5."
        msg += " MoleculePreparation.prepare() returns a list of MoleculeSetup instances."
        warnings.warn(msg, DeprecationWarning)
        return self.deprecated_setup_access

    @classmethod
    def get_defaults_dict(cls):
        defaults = {}
        sig = signature(cls)
        for key in sig.parameters:
            defaults[key] = sig.parameters[key].default 
        return defaults

    @classmethod
    def from_config(cls, config):
        expected_keys = cls.get_defaults_dict().keys()
        bad_keys = [k for k in config if k not in expected_keys]
        if len(bad_keys) > 0:
            err_msg = "unexpected keys in MoleculePreparation.from_config():" + os.linesep
            for key in bad_keys:
                err_msg += "  - %s" % key + os.linesep
            raise ValueError(err_msg)
        p = cls(**config)
        return p

    def __call__(self, *args):
        return self.prepare(*args)

    def prepare(self,
            mol,
            root_atom_index=None,
            not_terminal_atoms=[],
            delete_ring_bonds=[],
            glue_pseudo_atoms={},
            conformer_id=-1,
        ):
        """ 
        Create molecule setup from RDKit molecule

        Args:
            mol (rdkit.Chem.rdchem.Mol): with explicit hydrogens and 3D coordinates
            root_atom_index (int): to set ROOT of torsion tree instead of searching
            not_terminal_atoms (list): make bonds with terminal atoms rotatable
                                       (e.g. C-Alpha carbon in flexres)
            delete_ring_bonds (list): bonds deleted for macrocycle flexibility
                                      each bond is a tuple of two ints (atom 0-indices)
            glue_pseudo_atoms (dict): keys are parent atom indices, values are (x, y, z)
        """
        mol_type = type(mol)
        if not mol_type in self._classes_setup:
            raise TypeError("Molecule is not an instance of supported types: %s" % type(mol))
        setup_class = self._classes_setup[mol_type]
        setup = setup_class.from_mol(mol,
            keep_chorded_rings=self.keep_chorded_rings,
            keep_equivalent_rings=self.keep_equivalent_rings,
            assign_charges=self.charge_model=="gasteiger",
            conformer_id=conformer_id,
            )

        self.check_external_ring_break(setup, delete_ring_bonds, glue_pseudo_atoms)

        # 1.  assign atom params
        AtomTyper.type_everything(
            setup,
            self.atom_params,
            self.charge_model,
            self.offatom_params,
            self.dihedral_params,
        )
        
        # Convert molecule to graph and apply trained Espaloma model
        if self.dihedral_model == "espaloma" or self.charge_model == "espaloma":
            molgraph = self.espaloma_model.get_espaloma_graph(setup)

        # Grab dihedrals from graph node and set them to the molsetup
        if self.dihedral_model == "espaloma":
            self.espaloma_model.set_espaloma_dihedrals(setup, molgraph)

        # Grab charges from graph node and set them to the molsetup
        if self.charge_model == "espaloma":
            self.espaloma_model.set_espaloma_charges(setup, molgraph)

        # merge hydrogens (or any terminal atoms)
        indices = set()
        for atype_to_merge in self.merge_these_atom_types:
            for index, atype in setup.atom_type.items():
                if atype == atype_to_merge:
                    indices.add(index)
        setup.merge_terminal_atoms(indices)

        # 3.  assign bond types by using SMARTS...
        #     - bonds should be typed even in rings (but set as non-rotatable)
        #     - if macrocycle is selected, they will be enabled (so they must be typed already!)
        self._bond_typer(setup, self.flexible_amides, self.rigidify_bonds_smarts, self.rigidify_bonds_indices, not_terminal_atoms)
        # 4 . hydrate molecule
        if self.hydrate:
            self._water_builder.hydrate(setup)
        # 5.  break macrocycles into open/linear form
        if self.rigid_macrocycles:
            break_combo_data = None
            bonds_in_rigid_rings = None # not true, but this is only needed when breaking macrocycles
        else:
            break_combo_data, bonds_in_rigid_rings = self._macrocycle_typer.search_macrocycle(setup, delete_ring_bonds)

        # 6.  build flexibility...
        # 6.1 if macrocycles typed:
        #     - walk the setup graph by skipping proposed closures
        #       and score resulting flex_trees basing on the lenght
        #       of the branches generated
        #     - actually break the best closure bond (THIS CHANGES SOME ATOM TYPES)
        # 6.2  - walk the graph and build the flextree
        # 7.  but disable all bonds that are in rings and not
        #     in flexible macrocycles
        # TODO restore legacy AD types for PDBQT
        #self._atom_typer.set_param_legacy(mol)

        setup = self._flex_builder(setup,
                                   root_atom_index=root_atom_index,
                                   break_combo_data=break_combo_data,
                                   bonds_in_rigid_rings=bonds_in_rigid_rings,
                                   glue_pseudo_atoms=glue_pseudo_atoms,
        )

        if self.reactive_smarts is None:
            setups = [setup]
        else:
            reactive_types_dicts = assign_reactive_types(
                    setup,
                    self.reactive_smarts,
                    self.reactive_smarts_idx,
            )
            setups = []
            for r in reactive_types_dicts:
                new_setup = setup.copy()
                new_setup.atom_type = r
                setups.append(new_setup)

        self.deprecated_setup_access = setups[0] # for a gentle introduction of the new API
        return setups


    @staticmethod
    def check_external_ring_break(molsetup, break_ring_bonds, glue_pseudo_atoms):
        for (index1, index2) in break_ring_bonds:
            has_bond = molsetup.get_bond_id(index1, index2) in molsetup.bond
            if not has_bond:
                raise ValueError("bond (%d, %d) not in molsetup" % (index1, index2))
            for index in (index1, index2):
                if index not in glue_pseudo_atoms:
                    raise ValueError("missing glue pseudo for atom %d" % index) 
                xyz = glue_pseudo_atoms[index]
                if len(xyz) != 3:
                    raise ValueError("expected 3 coordinates (got %d) for glue pseudo of atom %d" % (len(xyz), index)) 


    def write_pdbqt_string(self, add_index_map=None, remove_smiles=None):
        msg = "MoleculePreparation.write_pdbqt_string() is deprecated in Meeko v0.5."
        msg += " Pass the MoleculeSetup instance to PDBQTWriterLegacy.write_string()."
        msg += " MoleculePreparation.prepare() returns a list of MoleculeSetup instances."
        warnings.warn(msg, DeprecationWarning)
        pdbqt_string, is_ok, err_msg = PDBQTWriterLegacy.write_string(self.setup)
        if not is_ok:
            msg = 'Cannot generate PDBQT, error from PDBQTWriterLegacy:' + os.linesep
            msg += err_msg
            raise RuntimeError(msg)
        return pdbqt_string


    def write_pdbqt_file(self, pdbqt_filename, add_index_map=None, remove_smiles=None):
        warnings.warn("MoleculePreparation.write_pdbqt_file() is deprecated since Meeko v0.5", DeprecationWarning)
        with open(pdbqt_filename,'w') as w:
            w.write(self.write_pdbqt_string(add_index_map, remove_smiles))




from rdkit import Chem
from .utils import mini_periodic_table


"""
create new RDKIT residue

mi  =  Chem.AtomPDBResidueInfo()
mi.SetResidueName('MOL')
mi.SetResidueNumber(1)
mi.SetOccupancy(0.0)
mi.SetTempFactor(0.0)

source: https://sourceforge.net/p/rdkit/mailman/message/36404394/
"""

from collections import namedtuple
PDBAtomInfo = namedtuple('PDBAtomInfo', "name resName resNum chain")

def getPdbInfoNoNull(atom):
    """extract information for populating an ATOM/HETATM line
    in the PDB"""
    # res = atom.GetResidue()
    minfo = atom.GetMonomerInfo()
    if minfo is None:
        atomic_number = atom.GetAtomicNum()
        if atomic_number == 0:
            name = '%-2s' % '*'
        else:
            name = '%-2s' % mini_periodic_table[atomic_number]
        chain = ' '
        resNum = 1
        resName = 'UNL'
    else:
        name = minfo.GetName()
        chain = minfo.GetChainId()
        resNum = minfo.GetResidueNumber()
        resName = minfo.GetResidueName()
    return PDBAtomInfo(name=name, resName=resName, resNum=resNum, chain=chain)


class Mol2MolSupplier():
    """ RDKit Mol2 molecule supplier.
    Parameters
        sanitize: perform RDKit sanitization of Mol2 molecule"""
    def __init__(self, filename, sanitize=True, removeHs=False, cleanupSubstructures=True):
        self.fp = open(filename, 'r')
        self._opts = {'sanitize':sanitize,
                'removeHs':removeHs,
                'cleanupSubstructures':cleanupSubstructures }
        self.buff = []

    def __iter__(self):
        return self

    def __next__(self):
        """ iterator step """
        while True:
            line = self.fp.readline()
            # empty line
            if not line:
                if len(self.buff):
                    # buffer full, returning last molecule
                    mol=Chem.MolFromMol2Block("".join(self.buff), **self._opts)
                    self.buff = []
                    return mol
                # buffer empty, stopping the iteration
                self.fp.close()
                raise StopIteration
            if '@<TRIPOS>MOLECULE' in line:
                # first molecule parsed
                if len(self.buff)==0:
                    self.buff.append(line)
                else:
                    # found the next molecule, breaking to return the complete one
                    break
            else:
                # adding another line in the current molecule
                self.buff.append(line)
        # found a complete molecule, returning it
        mol=Chem.MolFromMol2Block("".join(self.buff), **self._opts)
        self.buff = [line]
        return mol

def react_and_map(reactants, rxn):
    """run reaction and keep track of atom indices from reagents to products"""

    # https://github.com/rdkit/rdkit/discussions/4327#discussioncomment-998612

    # keep track of the reactant that each atom belongs to
    for i, reactant in enumerate(reactants):
        for atom in reactant.GetAtoms():
            atom.SetIntProp("reactant_idx", i)

    reactive_atoms_idxmap = {}
    for i in range(rxn.GetNumReactantTemplates()):
        reactant_template = rxn.GetReactantTemplate(i)
        for atom in reactant_template.GetAtoms():
            if atom.GetAtomMapNum():
                reactive_atoms_idxmap[atom.GetAtomMapNum()] = i

    products_list = rxn.RunReactants(reactants)

    index_map = {"reactant_idx": [], "atom_idx": []}
    for products in products_list:
        result_reac_map = []
        result_atom_map = []
        for product in products:
            atom_idxmap = [] 
            reac_idxmap = []
            for atom in product.GetAtoms():
                if atom.HasProp("reactant_idx"):
                    reactant_idx = atom.GetIntProp("reactant_idx")
                elif atom.HasProp("old_mapno"): # this atom matched the reaction SMARTS
                    reactant_idx = reactive_atoms_idxmap[atom.GetIntProp("old_mapno")]
                    #if atom.HasProp("reactant_idx"):
                    #    raise RuntimeError("did not expect both old_mapno and reactant_idx")
                    #    #print("did not expect both old_mapno and reactant_idx")
                else:
                    reactant_idx = None # the reaction SMARTS creates new atoms
                reac_idxmap.append(reactant_idx)
                if atom.HasProp("react_atom_idx"):
                    atom_idxmap.append(atom.GetIntProp("react_atom_idx"))
                else:
                    atom_idxmap.append(None)
            result_reac_map.append(reac_idxmap)
            result_atom_map.append(atom_idxmap)
        index_map["reactant_idx"].append(result_reac_map)
        index_map["atom_idx"].append(result_atom_map)

    return products_list, index_map



#!/usr/bin/env python

import json
import pathlib
import sys

from rdkit import Chem
import openforcefields

from meeko import MoleculePreparation
from meeko import get_openff_epsilon_sigma
from meeko import load_openff

vdw_list, _, vdw_by_type = load_openff()
sdf_filename = sys.argv[1]
rdkit_mol = Chem.MolFromMolFile(sdf_filename, removeHs=False)
data = get_openff_epsilon_sigma(
    rdkit_mol,
    vdw_list,
    vdw_by_type,
    output_index_start=0,
)
print(json.dumps(data, indent=4))
    

